#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Context classes
============================================================================

Dragonfly uses context classes to define when grammars and
rules should be active.  A context is an object with a
:meth:`Context.matches` method which returns *True* if the
system is currently within that context, and *False* if it
is not.

The following context classes are available:

 - :class:`Context` --
   the base class from which all other context classes are derived
 - :class:`AppContext` --
   class which is based on the application context, i.e. foreground
   window executable, title, and handle
 - :class:`FuncContext` --
   class that evaluates a given function/lambda/callable, whose return
   value is interpreted as a *bool*, determining whether the context is
   active


Logical operations
----------------------------------------------------------------------------

It is possible to modify and combine the behavior of contexts using the
Python's standard logical operators:

:logical AND: ``context1 & context2`` -- *all* contexts must match
:logical OR: ``context1 | context2`` --
   *one or more* of the contexts must match
:logical NOT: ``~context1`` -- *inversion* of when the context matches

For example, to create a context which will match when
Firefox is in the foreground, but only if Google Reader is
*not* being viewed::

   firefox_context = AppContext(executable="firefox")
   reader_context = AppContext(executable="firefox", title="Google Reader")
   firefox_but_not_reader_context = firefox_context & ~reader_context


Matching other window attributes
----------------------------------------------------------------------------

The :class:`AppContext` class can be used to match window attributes and
properties other than the title and executable. To do this, pass extra
keyword arguments to the constructor::

   # Context for a maximized Firefox window.
   maximized_firefox = AppContext(executable="firefox", is_maximized=True)

   # Context for a browser in fullscreen mode.
   # 'role' and 'is_fullscreen' are X11 only.
   fullscreen_browser = AppContext(role="browser", is_fullscreen=True)

   # Context for Android Studio or PyCharm using the X11 'cls' property.
   AppContext(cls=["jetbrains-studio", "jetbrains-pycharm-ce"])



Class reference
----------------------------------------------------------------------------

"""

import copy
import inspect
import logging


# --------------------------------------------------------------------------
from six import string_types


class Context(object):
    """
        Base class for other context classes.

        This base class implements some basic
        infrastructure, including what's required for
        logical operations on context objects.  Derived
        classes should at least do the following things:

         - During initialization, set *self._str* to some descriptive,
           human readable value.  This attribute is used by the
           ``__str__()`` method.
         - Overload the :meth:`Context.matches` method to implement
           the logic to determine when to be active.

        The *self._log* logger objects should be used in methods of
        derived classes for logging purposes.  It is a standard logger
        object from the *logger* module in the Python standard library.

    """

    _log = logging.getLogger("context.match")
    _log_match = _log

    # ----------------------------------------------------------------------
    # Initialization and aggregation methods.

    def __init__(self):
        self._str = ""

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self._str)

    def copy(self):
        return copy.deepcopy(self)

    # ----------------------------------------------------------------------
    # Logical operations.

    def __and__(self, other):
        return LogicAndContext(self, other)

    def __or__(self, other):
        return LogicOrContext(self, other)

    def __invert__(self):
        return LogicNotContext(self)

    # ----------------------------------------------------------------------
    # Matching methods.

    def matches(self, executable, title, handle):
        """
            Indicate whether the system is currently within this context.

            Arguments:
             - *executable* (*str*) --
               path name to the executable of the foreground application
             - *title* (*str*) -- title of the foreground window
             - *handle* (*int*) -- window handle to the foreground window

            The default implementation of this method simply returns *True*.

            .. note::

               This is generally the method which developers should
               overload to give derived context classes custom
               functionality.

        """
        # pylint: disable=unused-argument,no-self-use
        return True


# --------------------------------------------------------------------------
# Wrapper contexts for combining contexts in logical structures.

class LogicAndContext(Context):

    def __init__(self, *children):
        Context.__init__(self)
        self._children = children
        self._str = ", ".join(str(child) for child in children)

    def matches(self, executable, title, handle):
        for child in self._children:
            if not child.matches(executable, title, handle):
                return False
        return True


class LogicOrContext(Context):

    def __init__(self, *children):
        Context.__init__(self)
        self._children = children
        self._str = ", ".join(str(child) for child in children)

    def matches(self, executable, title, handle):
        for child in self._children:
            if child.matches(executable, title, handle):
                return True
        return False


class LogicNotContext(Context):

    def __init__(self, child):
        Context.__init__(self)
        self._child = child
        self._str = str(child)

    def matches(self, executable, title, handle):
        return not self._child.matches(executable, title, handle)


# --------------------------------------------------------------------------

class AppContext(Context):
    """
        Context class using foreground application details.

        This class determines whether the foreground window meets
        certain requirements.  Which requirements must be met for this
        context to match are determined by the constructor arguments.

        If multiple strings are passed in a list, True will be returned if
        the foreground window matches one or more of them. This applies to
        the *executable* and *title* arguments and key word arguments for
        most window attributes.

        Constructor arguments:
         - *executable* (*str* or *list*) --
           (part of) the path name of the foreground application's
           executable; case insensitive
         - *title* (*str* or *list*) --
           (part of) the title of the foreground window; case insensitive
         - *key word arguments* --
           optional window attributes/properties and expected values;
           case insensitive

    """

    # ----------------------------------------------------------------------
    # Initialization methods.

    def __init__(self, executable=None, title=None, exclude=False,
                 **kwargs):
        # pylint: disable=too-many-branches
        # Suppress warnings about too many if-else branches.
        Context.__init__(self)

        # Allow Unicode or strings to be used for executables and titles.
        if isinstance(executable, string_types):
            self._executable = [executable.lower()]
        elif isinstance(executable, (list, tuple)):
            self._executable = [e.lower() for e in executable]
        elif executable is None:
            self._executable = None
        else:
            raise TypeError("executable argument must be a string or None;"
                            " received %r" % executable)

        if isinstance(title, string_types):
            self._title = [title.lower()]
        elif isinstance(title, (list, tuple)):
            self._title = [t.lower() for t in title]
        elif title is None:
            self._title = None
        else:
            raise TypeError("title argument must be a string or None;"
                            " received %r" % title)

        # Handle keyword arguments.
        new_kwargs = {}
        for key, value in kwargs.items():
            if isinstance(value, string_types):
                values = [value.lower()]
            elif isinstance(value, list):
                values = [str(v).lower() for v in value]
            elif value is None:
                values = None
            else:
                values = [value]
            new_kwargs[key] = values

        self._exclude = bool(exclude)
        self._str = "%s, %s, %s" % (self._executable, self._title,
                                    self._exclude)
        self._kwargs = new_kwargs
        if self._kwargs:
            self._str += ", %s" % self._kwargs

    # ----------------------------------------------------------------------
    # Matching methods.

    def matches(self, executable, title, handle):
        # pylint: disable=too-many-branches
        # Suppress warnings about too many if-else branches.
        if isinstance(executable, string_types):
            executable = executable.lower()
        if isinstance(title, string_types):
            title = title.lower()

        if self._executable:
            found = False
            if isinstance(executable, string_types):
                for match in self._executable:
                    if executable.find(match) != -1:
                        found = True
                        break
            if self._exclude == found:
                self._log_match.debug("%s: No match, executable doesn't "
                                      "match.", self)
                return False

        if self._title:
            found = False
            if isinstance(title, string_types):
                for match in self._title:
                    if title.find(match) != -1:
                        found = True
                        break
            if self._exclude == found:
                self._log_match.debug("%s: No match, title doesn't match.",
                                      self)
                return False

        if self._kwargs:
            # Import locally to avoid import cycles.
            from dragonfly.windows import Window
            window = Window.get_window(handle)

            found = False
            for attr, expected_values in self._kwargs.items():
                try:
                    # Get the window attribute.
                    attr_value = getattr(window, attr)
                except AttributeError:
                    self._log_match.warning("%s: Skipped missing window"
                                            " attribute '%s'",
                                            self, attr)
                    continue

                # Check if the window attribute matched.
                for match in expected_values:
                    is_string = isinstance(attr_value, string_types)
                    found = (
                        not is_string and match == attr_value or
                        is_string and attr_value.lower().find(match) != -1
                    )
                    if found:
                        break

            if self._exclude == found:
                self._log_match.debug("%s: No match, not all extra"
                                      " attributes match.", self)
                return False

        if self._log_match:
            self._log_match.debug("%s: Match.", self)
        return True


# --------------------------------------------------------------------------

class FuncContext(Context):
    """
        Context class that evaluates a given function, whose return
        value is interpreted as a *bool*, determining whether the
        context is active.

        The foreground application details are optionally passed to the
        function as arguments named *executable*, *title*, and/or
        *handle*, if any/each matches a so-named keyword argument of the
        function. Default arguments may also be passed to the function,
        through this class's constructor.

    """

    def __init__(self, function, **defaults):
        """
            Constructor arguments:
             - *function* (callable) --
               the function to call when this context is evaluated
             - defaults --
               optional default keyword-values for the arguments with
               which the function will be called

        """

        Context.__init__(self)
        self._function = function
        self._defaults = defaults
        self._str = "%s, defaults: %s" % (self._function, self._defaults)

        (args, _, varkw, defaults) = inspect.getargspec(self._function)
        if varkw:  self._filter_keywords = False
        else:      self._filter_keywords = True
        self._valid_keywords = set(args)

    def matches(self, executable, title, handle):
        arguments = dict(self._defaults)
        arguments.update(executable=executable, title=title, handle=handle)

        if self._filter_keywords:
            invalid_keywords = set(arguments.keys()) - self._valid_keywords
            for key in invalid_keywords:
                del arguments[key]

        try:
            match = bool(self._function(**arguments))
            if match:
                self._log_match.debug("%s: Match.", self)
            else:
                self._log_match.debug("%s: No match, function"
                                      " returned false.", self)
            return match
        except:
            self._log.exception("Exception from function %s:",
                                self._function.__name__)
            # Fallback to matching
            return True



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
    This file implements the State class, a utility
    class used during recognition decoding.
"""

# pylint: disable=too-many-instance-attributes,no-self-use

from locale import getpreferredencoding

from logging import getLogger, DEBUG

from six import PY2, text_type, binary_type

from ..error import GrammarError


class State(object):

    _log_decode = getLogger("grammar.decode")

    # -----------------------------------------------------------------------
    # Methods for initialization.

    def __init__(self, results, rule_names, engine):
        self._results = results
        self._rule_names = rule_names
        self._engine = engine
        self._index = 0
        self._data = {}
        self._depth = 0
        self._stack = []
        self.initialize_decoding()
        self._previous_index = None

    def __repr__(self):
        if PY2:
            return self.__unicode__().encode(getpreferredencoding())
        else:
            return self.__unicode__()

    def __unicode__(self):
        words = self.words()
        before = words[:self._index]
        after = words[self._index:]
        return u" ".join(before) + u" >> " + u" ".join(after)

    # -----------------------------------------------------------------------
    # Methods for accessing recognition content.

    def word(self, delta=0):
        i = self._index + delta
        if 0 <= i < len(self._results):
            return self._results[i][0]
        else:
            return None

    @property
    def engine(self):
        return self._engine

    def rule(self, delta=0):
        i = self._index + delta
        if 0 <= i < len(self._results):
            rule_id = self._results[i][1]
            if 0 <= rule_id < len(self._rule_names):
                return self._rule_names[rule_id]
            elif rule_id == 1000000:
                return "dgndictation"
            elif rule_id == 1000001:
                return "dgnletters"
            else:
                word = self._results[i][0]
                raise GrammarError("Malformed recognition data:"
                                   " word %r, rule id %d."
                                   % (word, rule_id))
        else:
            return None

    def word_rule(self, delta=0):
        i = self._index + delta
        if 0 <= i < len(self._results):
            return self._results[i][0:2]
        else:
            return None

    def words(self, begin=0, end=None):
        return [w[0] for w in self._results[begin:end]]

    def next(self, delta=1):
        self._index += delta

    def finished(self):
        return self._index >= len(self._results)

    # -----------------------------------------------------------------------
    # Methods for tracking decoding of recognition.

    class Frame(object):
        __slots__ = ("depth", "actor", "begin", "end")

        def __init__(self, depth, actor, begin):
            self.depth = depth
            self.actor = actor
            self.begin = begin
            self.end = None

    def initialize_decoding(self):
        self._depth = 0
        self._stack = []
        self._previous_index = None

    def decode_attempt(self, element):
        self._depth += 1
        self._stack.append(State.Frame(self._depth, element, self._index))
        self._log_step(element, "attempt")

    def decode_retry(self, element):
        frame = self._get_frame_from_actor(element)
        self._depth = frame.depth
        self._log_step(element, "retry")

    def decode_rollback(self, element):
        frame = self._get_frame_from_depth()
        if not frame or frame.actor != element:
            raise GrammarError("Recognition decoding stack broken")
        if frame is self._stack[-1]:
            # Last parser on the stack, rollback.
            self._index = frame.begin
        else:
            raise GrammarError("Recognition decoding stack broken")
        self._log_step(element, "rollback")

    def decode_success(self, element):
        self._log_step(element, "success")
        frame = self._get_frame_from_depth()
        if not frame or frame.actor != element:
            raise GrammarError("Recognition decoding stack broken.")
        frame.end = self._index
        self._depth -= 1

    def decode_failure(self, element):
        frame = self._stack.pop()
        self._index = frame.begin
        self._depth = frame.depth
        self._log_step(element, "failure")
        self._depth -= 1

    def _get_frame_from_depth(self):
        for i in range(len(self._stack)-1, -1, -1):
            frame = self._stack[i]
            if frame.depth == self._depth:
                return frame
        return None

    def _get_frame_from_actor(self, actor):
        for i in range(len(self._stack)-1, -1, -1):
            frame = self._stack[i]
            if frame.actor is actor:
                return frame
        return None

    def _log_step(self, parser, message):
        if not self._log_decode or not self._log_decode.isEnabledFor(DEBUG):
            return
        indent = u"   " * self._depth
        output = u"%s%s: %s" % (indent, message, parser)
        if isinstance(output, binary_type):
            output = output.decode(getpreferredencoding())
        self._log_decode.debug(output)
        if self._index != self._previous_index:
            self._previous_index = self._index
            output = u"%s -- Decoding State: '%s'" % (indent, text_type(self))
            if isinstance(output, binary_type):
                output = output.decode(getpreferredencoding())
            self._log_decode.debug(output)

    # -----------------------------------------------------------------------
    # Methods for evaluation.

    def build_parse_tree(self):

        root = None
        node = None
        for frame in self._stack:
            while node and node.depth >= frame.depth:
                node = node.parent
            parent = node
            node = Node(parent, frame.actor, self._results,
                        frame.begin, frame.end, frame.depth, self._engine)
            if parent:
                parent.children.append(node)
            else:
                root = node

        return root


# ---------------------------------------------------------------------------

class Node(object):

    __slots__ = ("parent", "children", "actor", "results",
                 "begin", "end", "depth", "engine")

    # pylint: disable=too-many-arguments
    def __init__(self, parent, actor, results, begin, end, depth, engine):
        self.parent = parent
        self.actor = actor
        self.results = results
        self.begin = begin
        self.end = end
        self.depth = depth
        self.engine = engine
        self.children = []

    def __repr__(self):
        return "Node: %s, %s" % (self.actor, self.words())

    def words(self):
        return [w[0] for w in self.results[self.begin:self.end]]

    def full_results(self):
        return self.results[self.begin:self.end]

    def value(self):
        return self.actor.value(self)

    def pretty_string(self, indent=""):
        if not self.children:
            return "%s%s -> %r" % (indent, str(self), self.value())
        else:
            return "%s%s -> %r\n" % (indent, str(self), self.value()) \
                + "\n".join([n  .pretty_string(indent + "  ")
                             for n in self.children])

    def _get_name(self):
        return self.actor.name
    name = property(_get_name)

    def has_child_with_name(self, name):
        """True if at least one node below this node has the given name."""
        for child in self.children:
            if child.name == name:
                return True
            if child.has_child_with_name(name):
                return True
        return False

    def get_child_by_name(self, name, shallow=False):
        """Get one node below this node with the given name."""
        for child in self.children:
            if child.name:
                if child.name == name:
                    return child
                if shallow:
                    # If shallow, don't look past named children.
                    continue
            match = child.get_child_by_name(name, shallow)
            if match:
                return match
        return None

    def get_children_by_name(self, name, shallow=False):
        """
        Get all nodes below this node with the given name.
        """
        matches = []
        for child in self.children:
            if child.name:
                if child.name == name:
                    matches.append(child)
                if shallow:
                    # If shallow, don't look past named children.
                    continue
            matches.extend(child.get_children_by_name(name, shallow))
        return matches



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
English language implementations of Integer and Digits classes
============================================================================

"""

from ..base.integer_internal  import (MapIntBuilder, CollectionIntBuilder,
                                      MagnitudeIntBuilder, IntegerContentBase)
from ..base.digits_internal   import DigitsContentBase


#---------------------------------------------------------------------------

int_0           = MapIntBuilder({
                                 "(zero | oh)":         0,
                               })
int_1_9         = MapIntBuilder({
                                 "one":                 1,
                                 "two":                 2,
                                 "three":               3,
                                 "four":                4,
                                 "five":                5,
                                 "six":                 6,
                                 "seven":               7,
                                 "eight":               8,
                                 "nine":                9,
                               })
int_10_19       = MapIntBuilder({
                                 "ten":       10,
                                 "eleven":    11,
                                 "twelve":    12,
                                 "thirteen":  13,
                                 "fourteen":  14,
                                 "fifteen":   15,
                                 "sixteen":   16,
                                 "seventeen": 17,
                                 "eighteen":  18,
                                 "nineteen":  19,
                               })
int_20_90_10    = MapIntBuilder({
                                 "twenty":     2,
                                 "thirty":     3,
                                 "forty":      4,
                                 "fifty":      5,
                                 "sixty":      6,
                                 "seventy":    7,
                                 "eighty":     8,
                                 "ninety":     9,
                               })
int_20_99       = MagnitudeIntBuilder(
                   factor      = 10,
                   spec        = "<multiplier> [<remainder>]",
                   multipliers = [int_20_90_10],
                   remainders  = [int_1_9],
                  )
int_and_1_99    = CollectionIntBuilder(
                   spec        = "[and] <element>",
                   set         = [int_1_9, int_10_19, int_20_99],
                  )
int_100s        = MagnitudeIntBuilder(
                   factor      = 100,
                   spec        = "[<multiplier>] hundred [<remainder>]",
                   multipliers = [int_1_9],
                   remainders  = [int_and_1_99],
                  )
int_100big      = MagnitudeIntBuilder(
                   factor      = 100,
                   spec        = "[<multiplier>] hundred [<remainder>]",
                   multipliers = [int_10_19, int_20_99],
                   remainders  = [int_and_1_99]
                  )
int_1000s       = MagnitudeIntBuilder(
                   factor      = 1000,
                   spec        = "[<multiplier>] thousand [<remainder>]",
                   multipliers = [int_1_9, int_10_19, int_20_99, int_100s],
                   remainders  = [int_and_1_99, int_100s]
                  )
int_1000000s    = MagnitudeIntBuilder(
                   factor      = 1000000,
                   spec        = "[<multiplier>] million [<remainder>]",
                   multipliers = [int_1_9, int_10_19, int_20_99, int_100s, int_1000s],
                   remainders  = [int_and_1_99, int_100s, int_1000s],
                  )


#---------------------------------------------------------------------------

class IntegerContent(IntegerContentBase):
    builders = [int_0, int_1_9, int_10_19, int_20_99,
                int_100s, int_100big, int_1000s, int_1000000s]

class DigitsContent(DigitsContentBase):
    digits = [("zero", "oh"), "one", ("two", "too", "to"), "three",
              "four", "five", "six", "seven", "eight", "nine"]



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Win32 input wrapper functions.

This file implements an interface to the Win32 SendInput function
for simulating keyboard and mouse events.
"""

# pylint: disable=E0401
# This file imports Win32-only symbols.

import sys
from ctypes import (c_short, c_long, c_ushort, c_ulong, sizeof,
                    POINTER, pointer, Structure, Union, windll)

import win32con
import win32api


class KeyboardInput(Structure):
    """Win32 KEYBDINPUT wrapper."""

    _fields_ = [("wVk", c_ushort),
                ("wScan", c_ushort),
                ("dwFlags", c_ulong),
                ("time", c_ulong),
                ("dwExtraInfo", POINTER(c_ulong))]

    #  From https://docs.microsoft.com/en-us/windows/desktop/inputdev/about-keyboard-input#extended-key-flag
    #     The extended keys consist of the ALT and CTRL keys
    #     on the right-hand side of the keyboard; the INS, DEL, HOME,
    #     END, PAGE UP, PAGE DOWN, and arrow keys in the clusters to
    #     the left of the numeric keypad; the NUM LOCK key; the BREAK
    #     (CTRL+PAUSE) key; the PRINT SCRN key; and the divide (/) and
    #     ENTER keys in the numeric keypad.
    #
    #  It's unclear if the Windows keys are also "extended", so they
    #  have been included for historical reasons.
    extended_keys = {
        win32con.VK_UP,
        win32con.VK_DOWN,
        win32con.VK_LEFT,
        win32con.VK_RIGHT,
        win32con.VK_HOME,
        win32con.VK_END,
        win32con.VK_PRIOR,
        win32con.VK_NEXT,
        win32con.VK_INSERT,
        win32con.VK_DELETE,
        win32con.VK_NUMLOCK,
        win32con.VK_RCONTROL,
        win32con.VK_RMENU,
        win32con.VK_PAUSE,
        win32con.VK_SNAPSHOT,
        win32con.VK_DIVIDE,
        win32con.VK_LWIN,
        win32con.VK_RWIN,
    }
    soft_keys = {
        win32con.VK_PAUSE,
    }


    def __init__(self, virtual_keycode, down, scancode=None, layout=None):
        """Initialize structure based on key type."""
        flags = 0
        if virtual_keycode == 0:
            flags |= 4  # KEYEVENTF_UNICODE
        else:
            # Translate *virtual_keycode* into a scan code, if necessary.
            #  Assume that *layout* is a keyboard layout (HKL).
            if scancode is None:
                user32 = windll.user32
                if sys.getwindowsversion().major < 6:
                    map_type = 0  # MAPVK_VK_TO_VSC
                else:
                    map_type = 4  # MAPVK_VK_TO_VSC_EX
                if layout is None:
                    scancode = user32.MapVirtualKeyW(virtual_keycode,
                                                     map_type)
                else:
                    scancode = user32.MapVirtualKeyExW(virtual_keycode,
                                                       map_type,
                                                       layout)

            # Add the KEYEVENTF_SCANCODE flag if the Win32 MapVirtualKey(Ex)
            #  function returned a translation.  If not, then fallback on
            #  the specified keycode.  This is also done for "soft" keys.
            if scancode and virtual_keycode not in self.soft_keys:
                flags |= 8  # KEYEVENTF_SCANCODE

                # Add the KEYEVENTF_EXTENDEDKEY flag, if necessary.
                #  Starting with Windows Vista, extended scan codes are
                #  specified with the high byte containing either 0xe0 or
                #  0xe1.
                if scancode >> 8 in (0xe0, 0xe1):
                    flags |= win32con.KEYEVENTF_EXTENDEDKEY

            # Always add the KEYEVENTF_EXTENDEDKEY flag for certain
            #  virtual-keys (see above).
            if virtual_keycode in self.extended_keys:
                flags |= win32con.KEYEVENTF_EXTENDEDKEY

        if not down:
            flags |= win32con.KEYEVENTF_KEYUP

        extra = pointer(c_ulong(0))
        Structure.__init__(self, virtual_keycode, scancode, flags, 0, extra)


class HardwareInput(Structure):
    _fields_ = [("uMsg", c_ulong),
                ("wParamL", c_short),
                ("wParamH", c_ushort)]

class MouseInput(Structure):
    _fields_ = [("dx", c_long),
                ("dy", c_long),
                ("mouseData", c_ulong),
                ("dwFlags", c_ulong),
                ("time",c_ulong),
                ("dwExtraInfo", POINTER(c_ulong))]


class _InputUnion(Union):
    _fields_ = [("ki", KeyboardInput),
                ("mi", MouseInput),
                ("hi", HardwareInput)]

class _Input(Structure):
    _fields_ = [("type", c_ulong),
                ("ii", _InputUnion)]

    def __init__(self, element):
        if   isinstance(element, KeyboardInput):
            element_type = win32con.INPUT_KEYBOARD
            union = _InputUnion(ki=element)
        elif isinstance(element, MouseInput):
            element_type = win32con.INPUT_MOUSE
            union = _InputUnion(mi=element)
        elif isinstance(element, HardwareInput):
            element_type = win32con.INPUT_HARDWARE
            union = _InputUnion(hi=element)
        else: raise TypeError("Unknown input type: %r" % element)

        Structure.__init__(self, type=element_type, ii=union)


def make_input_array(inputs):
    arguments = [(i,) for i in inputs]
    InputArray = _Input * len(inputs)
    return InputArray(*arguments)


def send_input_array(input_array):
    length = len(input_array)
    assert length >= 0
    size = sizeof(input_array[0])
    ptr = pointer(input_array)

    count_inserted = windll.user32.SendInput(length, ptr, size)

    if count_inserted != length:
        last_error = win32api.GetLastError()
        message = win32api.FormatMessage(last_error)
        raise ValueError("windll.user32.SendInput(): %s" % (message))



"""
Command-module loader for CMU Pocket Sphinx.

This script is based on 'dfly-loader-wsr.py' written by Christo Butcher and
has been adapted to work with the Sphinx engine instead.

This script can be used to look for Dragonfly command-modules for use with
the CMU Pocket Sphinx engine. It scans the directory it's in and loads any
``_*.py`` it finds.
"""


# TODO Have a simple GUI for pausing, resuming, cancelling and stopping
# recognition, etc

import logging
import os.path
import sys

import six

from dragonfly import get_engine
from dragonfly.loader import CommandModuleDirectory
from dragonfly.log import setup_log

# --------------------------------------------------------------------------
# Set up basic logging.

setup_log()


# --------------------------------------------------------------------------
# Main event driving loop.

def main():
    try:
        path = os.path.dirname(__file__)
    except NameError:
        # The "__file__" name is not always available, for example
        # when this module is run from PythonWin.  In this case we
        # simply use the current working directory.
        path = os.getcwd()
        __file__ = os.path.join(path, "sphinx_module_loader.py")

    # Initialize the engine.
    engine = get_engine("sphinx")

    # Try to import the local engine configuration object first. If there isn't one,
    # use the default engine configuration.
    log = logging.getLogger("config")
    try:
        import config
        engine.config = config
        log.info("Using local engine configuration module 'config.py'")
    except ImportError:
        pass
    except Exception as e:
        # Log errors caught when setting the configuration.
        log.exception("Failed to set config using 'config.py': %s" % e)
        log.warning("Falling back to the default engine configuration")

    # You can also set any configuration options here instead of using a
    # config.py file. For example:
    # engine.config.START_ASLEEP = False

    # Call connect() now that the engine configuration is set.
    engine.connect()

    # Load grammars.
    directory = CommandModuleDirectory(path, excludes=[__file__])
    directory.load()

    # Define recognition callback functions.
    def on_begin():
        print("Speech start detected.")

    def on_recognition(words):
        message = u"Recognized: %s" % u" ".join(words)

        # This only seems to be an issue with Python 2.7 on Windows.
        if six.PY2:
            encoding = sys.stdout.encoding or "ascii"
            message = message.encode(encoding, errors='replace')
        print(message)

    def on_failure():
        print("Sorry, what was that?")

    # Start the engine's main recognition loop
    try:
        engine.do_recognition(on_begin, on_recognition, on_failure)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
    This file builds the mapping from key-name to Typeable instances.

    To make changes/additions to the mappings, use the _generate_typeables.py
    script to generate the list and copy-paste the relevant section into
    this file.

    The script also generates documentation for the :class:`Key` action.

"""

from .keyboard import keyboard, Typeable, KeySymbols


# --------------------------------------------------------------------------
# Mapping of name -> typeable.

typeables = {}
key_symbols = KeySymbols()


def _add_typeable(name, char):
    # Add a character to the typeables dictionary if it is typeable with
    # the current keyboard layout.
    try:
        typeables[name] = keyboard.get_typeable(char)
    except ValueError:
        # Errors or log messages will occur later if code attempts to use
        # missing typeables.
        pass


# Lowercase letter keys
_add_typeable(name="a", char='a')
_add_typeable(name="alpha", char='a')
_add_typeable(name="b", char='b')
_add_typeable(name="bravo", char='b')
_add_typeable(name="c", char='c')
_add_typeable(name="charlie", char='c')
_add_typeable(name="d", char='d')
_add_typeable(name="delta", char='d')
_add_typeable(name="e", char='e')
_add_typeable(name="echo", char='e')
_add_typeable(name="f", char='f')
_add_typeable(name="foxtrot", char='f')
_add_typeable(name="g", char='g')
_add_typeable(name="golf", char='g')
_add_typeable(name="h", char='h')
_add_typeable(name="hotel", char='h')
_add_typeable(name="i", char='i')
_add_typeable(name="india", char='i')
_add_typeable(name="j", char='j')
_add_typeable(name="juliet", char='j')
_add_typeable(name="k", char='k')
_add_typeable(name="kilo", char='k')
_add_typeable(name="l", char='l')
_add_typeable(name="lima", char='l')
_add_typeable(name="m", char='m')
_add_typeable(name="mike", char='m')
_add_typeable(name="n", char='n')
_add_typeable(name="november", char='n')
_add_typeable(name="o", char='o')
_add_typeable(name="oscar", char='o')
_add_typeable(name="p", char='p')
_add_typeable(name="papa", char='p')
_add_typeable(name="q", char='q')
_add_typeable(name="quebec", char='q')
_add_typeable(name="r", char='r')
_add_typeable(name="romeo", char='r')
_add_typeable(name="s", char='s')
_add_typeable(name="sierra", char='s')
_add_typeable(name="t", char='t')
_add_typeable(name="tango", char='t')
_add_typeable(name="u", char='u')
_add_typeable(name="uniform", char='u')
_add_typeable(name="v", char='v')
_add_typeable(name="victor", char='v')
_add_typeable(name="w", char='w')
_add_typeable(name="whisky", char='w')
_add_typeable(name="x", char='x')
_add_typeable(name="xray", char='x')
_add_typeable(name="y", char='y')
_add_typeable(name="yankee", char='y')
_add_typeable(name="z", char='z')
_add_typeable(name="zulu", char='z')

# Uppercase letter keys
_add_typeable(name="A", char='A')
_add_typeable(name="Alpha", char='A')
_add_typeable(name="B", char='B')
_add_typeable(name="Bravo", char='B')
_add_typeable(name="C", char='C')
_add_typeable(name="Charlie", char='C')
_add_typeable(name="D", char='D')
_add_typeable(name="Delta", char='D')
_add_typeable(name="E", char='E')
_add_typeable(name="Echo", char='E')
_add_typeable(name="F", char='F')
_add_typeable(name="Foxtrot", char='F')
_add_typeable(name="G", char='G')
_add_typeable(name="Golf", char='G')
_add_typeable(name="H", char='H')
_add_typeable(name="Hotel", char='H')
_add_typeable(name="I", char='I')
_add_typeable(name="India", char='I')
_add_typeable(name="J", char='J')
_add_typeable(name="Juliet", char='J')
_add_typeable(name="K", char='K')
_add_typeable(name="Kilo", char='K')
_add_typeable(name="L", char='L')
_add_typeable(name="Lima", char='L')
_add_typeable(name="M", char='M')
_add_typeable(name="Mike", char='M')
_add_typeable(name="N", char='N')
_add_typeable(name="November", char='N')
_add_typeable(name="O", char='O')
_add_typeable(name="Oscar", char='O')
_add_typeable(name="P", char='P')
_add_typeable(name="Papa", char='P')
_add_typeable(name="Q", char='Q')
_add_typeable(name="Quebec", char='Q')
_add_typeable(name="R", char='R')
_add_typeable(name="Romeo", char='R')
_add_typeable(name="S", char='S')
_add_typeable(name="Sierra", char='S')
_add_typeable(name="T", char='T')
_add_typeable(name="Tango", char='T')
_add_typeable(name="U", char='U')
_add_typeable(name="Uniform", char='U')
_add_typeable(name="V", char='V')
_add_typeable(name="Victor", char='V')
_add_typeable(name="W", char='W')
_add_typeable(name="Whisky", char='W')
_add_typeable(name="X", char='X')
_add_typeable(name="Xray", char='X')
_add_typeable(name="Y", char='Y')
_add_typeable(name="Yankee", char='Y')
_add_typeable(name="Z", char='Z')
_add_typeable(name="Zulu", char='Z')

# Number keys
_add_typeable(name="0", char='0')
_add_typeable(name="zero", char='0')
_add_typeable(name="1", char='1')
_add_typeable(name="one", char='1')
_add_typeable(name="2", char='2')
_add_typeable(name="two", char='2')
_add_typeable(name="3", char='3')
_add_typeable(name="three", char='3')
_add_typeable(name="4", char='4')
_add_typeable(name="four", char='4')
_add_typeable(name="5", char='5')
_add_typeable(name="five", char='5')
_add_typeable(name="6", char='6')
_add_typeable(name="six", char='6')
_add_typeable(name="7", char='7')
_add_typeable(name="seven", char='7')
_add_typeable(name="8", char='8')
_add_typeable(name="eight", char='8')
_add_typeable(name="9", char='9')
_add_typeable(name="nine", char='9')

# Symbol keys
# All symbols can be referred to by their printable representation.
# Reserved characters for the Key action spec -,:/ are Typeable objects,
# but Key requires use of the longer character names.
_add_typeable(name="!", char='!')
_add_typeable(name="bang", char='!')
_add_typeable(name="exclamation", char='!')
_add_typeable(name="@", char='@')
_add_typeable(name="at", char='@')
_add_typeable(name="#", char='#')
_add_typeable(name="hash", char='#')
_add_typeable(name="$", char='$')
_add_typeable(name="dollar", char='$')
_add_typeable(name="%", char='%')
_add_typeable(name="percent", char='%')
_add_typeable(name="^", char='^')
_add_typeable(name="caret", char='^')
_add_typeable(name="&", char='&')
_add_typeable(name="and", char='&')
_add_typeable(name="ampersand", char='&')
_add_typeable(name="*", char='*')
_add_typeable(name="star", char='*')
_add_typeable(name="asterisk", char='*')
_add_typeable(name="(", char='(')
_add_typeable(name="leftparen", char='(')
_add_typeable(name="lparen", char='(')
_add_typeable(name=")", char=')')
_add_typeable(name="rightparen", char=')')
_add_typeable(name="rparen", char=')')
_add_typeable(name="-", char='-')
_add_typeable(name="minus", char='-')
_add_typeable(name="hyphen", char='-')
_add_typeable(name="_", char='_')
_add_typeable(name="underscore", char='_')
_add_typeable(name="+", char='+')
_add_typeable(name="plus", char='+')
_add_typeable(name="`", char='`')
_add_typeable(name="backtick", char='`')
_add_typeable(name="~", char='~')
_add_typeable(name="tilde", char='~')
_add_typeable(name="[", char='[')
_add_typeable(name="leftbracket", char='[')
_add_typeable(name="lbracket", char='[')
_add_typeable(name="]", char=']')
_add_typeable(name="rightbracket", char=']')
_add_typeable(name="rbracket", char=']')
_add_typeable(name="{", char='{')
_add_typeable(name="leftbrace", char='{')
_add_typeable(name="lbrace", char='{')
_add_typeable(name="}", char='}')
_add_typeable(name="rightbrace", char='}')
_add_typeable(name="rbrace", char='}')
_add_typeable(name="\\", char='\\')
_add_typeable(name="backslash", char='\\')
_add_typeable(name="|", char='|')
_add_typeable(name="bar", char='|')
_add_typeable(name=":", char=':')
_add_typeable(name="colon", char=':')
_add_typeable(name=";", char=';')
_add_typeable(name="semicolon", char=';')
_add_typeable(name="\'", char='\'')
_add_typeable(name="apostrophe", char='\'')
_add_typeable(name="singlequote", char='\'')
_add_typeable(name="squote", char='\'')
_add_typeable(name="\"", char='\"')
_add_typeable(name="quote", char='\"')
_add_typeable(name="doublequote", char='\"')
_add_typeable(name="dquote", char='\"')
_add_typeable(name=",", char=',')
_add_typeable(name="comma", char=',')
_add_typeable(name=".", char='.')
_add_typeable(name="dot", char='.')
_add_typeable(name="/", char='/')
_add_typeable(name="slash", char='/')
_add_typeable(name="<", char='<')
_add_typeable(name="lessthan", char='<')
_add_typeable(name="leftangle", char='<')
_add_typeable(name="langle", char='<')
_add_typeable(name=">", char='>')
_add_typeable(name="greaterthan", char='>')
_add_typeable(name="rightangle", char='>')
_add_typeable(name="rangle", char='>')
_add_typeable(name="?", char='?')
_add_typeable(name="question", char='?')
_add_typeable(name="=", char='=')
_add_typeable(name="equal", char='=')
_add_typeable(name="equals", char='=')

typeables.update({
    # Whitespace and editing keys
    "enter":            Typeable(code=key_symbols.RETURN, name='enter'),
    "tab":              Typeable(code=key_symbols.TAB, name='tab'),
    "space":            Typeable(code=key_symbols.SPACE, name='space'),
    "backspace":        Typeable(code=key_symbols.BACK, name='backspace'),
    "delete":           Typeable(code=key_symbols.DELETE, name='delete'),
    "del":              Typeable(code=key_symbols.DELETE, name='del'),

    # Main modifier keys
    "shift":            Typeable(code=key_symbols.SHIFT, name='shift'),
    "control":          Typeable(code=key_symbols.CONTROL, name='control'),
    "ctrl":             Typeable(code=key_symbols.CONTROL, name='ctrl'),
    "alt":              Typeable(code=key_symbols.ALT, name='alt'),

    # Right modifier keys
    "rshift":           Typeable(code=key_symbols.RSHIFT, name='rshift'),
    "rcontrol":         Typeable(code=key_symbols.RCONTROL, name='rcontrol'),
    "rctrl":            Typeable(code=key_symbols.RCONTROL, name='rctrl'),
    "ralt":             Typeable(code=key_symbols.RALT, name='ralt'),

    # Special keys
    "escape":           Typeable(code=key_symbols.ESCAPE, name='escape'),
    "insert":           Typeable(code=key_symbols.INSERT, name='insert'),
    "pause":            Typeable(code=key_symbols.PAUSE, name='pause'),
    "win":              Typeable(code=key_symbols.LSUPER, name='win'),
    "rwin":             Typeable(code=key_symbols.RSUPER, name='rwin'),
    "apps":             Typeable(code=key_symbols.APPS, name='apps'),
    "popup":            Typeable(code=key_symbols.APPS, name='popup'),
    "snapshot":         Typeable(code=key_symbols.SNAPSHOT, name='snapshot'),
    "printscreen":      Typeable(code=key_symbols.SNAPSHOT, name='printscreen'),

    # Lock keys
    # win32api.GetKeyState(code) could be used to toggle lock keys sensibly
    # instead of using the up/down modifiers.
    "scrolllock":       Typeable(code=key_symbols.SCROLL_LOCK, name='scrolllock'),
    "numlock":          Typeable(code=key_symbols.NUM_LOCK, name='numlock'),
    "capslock":         Typeable(code=key_symbols.CAPS_LOCK, name='capslock'),

    # Navigation keys
    "up":               Typeable(code=key_symbols.UP, name='up'),
    "down":             Typeable(code=key_symbols.DOWN, name='down'),
    "left":             Typeable(code=key_symbols.LEFT, name='left'),
    "right":            Typeable(code=key_symbols.RIGHT, name='right'),
    "pageup":           Typeable(code=key_symbols.PAGE_UP, name='pageup'),
    "pgup":             Typeable(code=key_symbols.PAGE_UP, name='pgup'),
    "pagedown":         Typeable(code=key_symbols.PAGE_DOWN, name='pagedown'),
    "pgdown":           Typeable(code=key_symbols.PAGE_DOWN, name='pgdown'),
    "home":             Typeable(code=key_symbols.HOME, name='home'),
    "end":              Typeable(code=key_symbols.END, name='end'),

    # Number pad keys
    "npmul":            Typeable(code=key_symbols.MULTIPLY, name='npmul'),
    "npadd":            Typeable(code=key_symbols.ADD, name='npadd'),
    "npsep":            Typeable(code=key_symbols.SEPARATOR, name='npsep'),
    "npsub":            Typeable(code=key_symbols.SUBTRACT, name='npsub'),
    "npdec":            Typeable(code=key_symbols.DECIMAL, name='npdec'),
    "npdiv":            Typeable(code=key_symbols.DIVIDE, name='npdiv'),
    "numpad0":          Typeable(code=key_symbols.NUMPAD0, name='numpad0'),
    "np0":              Typeable(code=key_symbols.NUMPAD0, name='np0'),
    "numpad1":          Typeable(code=key_symbols.NUMPAD1, name='numpad1'),
    "np1":              Typeable(code=key_symbols.NUMPAD1, name='np1'),
    "numpad2":          Typeable(code=key_symbols.NUMPAD2, name='numpad2'),
    "np2":              Typeable(code=key_symbols.NUMPAD2, name='np2'),
    "numpad3":          Typeable(code=key_symbols.NUMPAD3, name='numpad3'),
    "np3":              Typeable(code=key_symbols.NUMPAD3, name='np3'),
    "numpad4":          Typeable(code=key_symbols.NUMPAD4, name='numpad4'),
    "np4":              Typeable(code=key_symbols.NUMPAD4, name='np4'),
    "numpad5":          Typeable(code=key_symbols.NUMPAD5, name='numpad5'),
    "np5":              Typeable(code=key_symbols.NUMPAD5, name='np5'),
    "numpad6":          Typeable(code=key_symbols.NUMPAD6, name='numpad6'),
    "np6":              Typeable(code=key_symbols.NUMPAD6, name='np6'),
    "numpad7":          Typeable(code=key_symbols.NUMPAD7, name='numpad7'),
    "np7":              Typeable(code=key_symbols.NUMPAD7, name='np7'),
    "numpad8":          Typeable(code=key_symbols.NUMPAD8, name='numpad8'),
    "np8":              Typeable(code=key_symbols.NUMPAD8, name='np8'),
    "numpad9":          Typeable(code=key_symbols.NUMPAD9, name='numpad9'),
    "np9":              Typeable(code=key_symbols.NUMPAD9, name='np9'),

    # Function keys
    "f1":               Typeable(code=key_symbols.F1, name='f1'),
    "f2":               Typeable(code=key_symbols.F2, name='f2'),
    "f3":               Typeable(code=key_symbols.F3, name='f3'),
    "f4":               Typeable(code=key_symbols.F4, name='f4'),
    "f5":               Typeable(code=key_symbols.F5, name='f5'),
    "f6":               Typeable(code=key_symbols.F6, name='f6'),
    "f7":               Typeable(code=key_symbols.F7, name='f7'),
    "f8":               Typeable(code=key_symbols.F8, name='f8'),
    "f9":               Typeable(code=key_symbols.F9, name='f9'),
    "f10":              Typeable(code=key_symbols.F10, name='f10'),
    "f11":              Typeable(code=key_symbols.F11, name='f11'),
    "f12":              Typeable(code=key_symbols.F12, name='f12'),
    "f13":              Typeable(code=key_symbols.F13, name='f13'),
    "f14":              Typeable(code=key_symbols.F14, name='f14'),
    "f15":              Typeable(code=key_symbols.F15, name='f15'),
    "f16":              Typeable(code=key_symbols.F16, name='f16'),
    "f17":              Typeable(code=key_symbols.F17, name='f17'),
    "f18":              Typeable(code=key_symbols.F18, name='f18'),
    "f19":              Typeable(code=key_symbols.F19, name='f19'),
    "f20":              Typeable(code=key_symbols.F20, name='f20'),
    "f21":              Typeable(code=key_symbols.F21, name='f21'),
    "f22":              Typeable(code=key_symbols.F22, name='f22'),
    "f23":              Typeable(code=key_symbols.F23, name='f23'),
    "f24":              Typeable(code=key_symbols.F24, name='f24'),

    # Multimedia keys
    "volumeup":         Typeable(code=key_symbols.VOLUME_UP, name='volumeup'),
    "volup":            Typeable(code=key_symbols.VOLUME_UP, name='volup'),
    "volumedown":       Typeable(code=key_symbols.VOLUME_DOWN, name='volumedown'),
    "voldown":          Typeable(code=key_symbols.VOLUME_DOWN, name='voldown'),
    "volumemute":       Typeable(code=key_symbols.VOLUME_MUTE, name='volumemute'),
    "volmute":          Typeable(code=key_symbols.VOLUME_MUTE, name='volmute'),
    "tracknext":        Typeable(code=key_symbols.MEDIA_NEXT_TRACK, name='tracknext'),
    "trackprev":        Typeable(code=key_symbols.MEDIA_PREV_TRACK, name='trackprev'),
    "playpause":        Typeable(code=key_symbols.MEDIA_PLAY_PAUSE, name='playpause'),
    "browserback":      Typeable(code=key_symbols.BROWSER_BACK, name='browserback'),
    "browserforward":   Typeable(code=key_symbols.BROWSER_FORWARD, name='browserforward'),
})



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#


import unittest
from dragonfly import Context, AppContext


class TestContextOperations(unittest.TestCase):

    def setUp(self):
        self.context_yes = Context()
        self.context_no = AppContext("never")

    def test_logic(self):
        self._check_match(self.context_yes,   True)
        self._check_match(self.context_no,    False)
        self._check_match(~self.context_yes,  False)
        self._check_match(~self.context_no,   True)
        self._check_match(self.context_yes & self.context_yes,   True)
        self._check_match(self.context_yes & self.context_no,    False)
        self._check_match(self.context_no  & self.context_yes,   False)
        self._check_match(self.context_no  & self.context_no,    False)
        self._check_match(self.context_yes | self.context_yes,   True)
        self._check_match(self.context_yes | self.context_no,    True)
        self._check_match(self.context_no  | self.context_yes,   True)
        self._check_match(self.context_no  | self.context_no,    False)
        self._check_match(self.context_yes & ~self.context_no,   True)
        self._check_match(self.context_yes & ~self.context_yes,  False)

    def _check_match(self, context, expected, executable="", title="",
                     handle=None):
        self.assertEqual(context.matches(executable, title, handle),
                         expected)


if __name__ == '__main__':
    unittest.main()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Main test runner script
============================================================================

This file is Dragonfly's main test script.  Running it will execute all
tests within the Dragonfly library.


Dependencies
----------------------------------------------------------------------------

This script uses several external libraries to collect, run, and analyze
the Dragonfly tests.  It requires the following external dependencies:

 * `Nose <http://somethingaboutorange.com/mrl/projects/nose/>`_ --
   an extension on unittest.
 * `Coverage <http://nedbatchelder.com/code/coverage/>`_ --
   a code coverage measuring tool.

"""

import sys
import os.path
import subprocess


#===========================================================================
# Test control functions for the various types of tests: nose, pep8, pylint.

def run_nose():
    """ Use Nose to collect and execute test cases within Dragonfly. """
    import nose.core
#    import coverage.control
    from pkg_resources import resource_filename

    # Determine directory in which to save coverage report.
    setup_path = os.path.abspath(resource_filename("dragonfly", "license.txt"))
    directory = os.path.dirname(os.path.dirname(setup_path))
    cover_dir = os.path.join(directory, "coverage")

    # Virtual commandline arguments to be given to nose.
    argv = [
            sys.argv[0],
            "--failure-detail",
#            "--with-doctest",
#            "--doctest-extension=doctest",
            directory,
           ]

    # Clear coverage history and start new coverage measurement.
#    coverage = coverage.control.coverage()
#    coverage.erase()
#    coverage.start()

    # Let nose run tests.
    nose.core.TestProgram(argv=argv, exit=False)

    # Save coverage data and generate HTML report.
#    coverage.stop()
#    coverage.save()
#    coverage.html_report(directory=cover_dir)


def run_pylint():
    """ Use the pylint package to analyze the Dragonfly source code. """
    from pylint import lint
    args = [
            "dragonfly",
           ]
    lint.Run(args)


def run_pep8():
    """ Use the pep8 package to analyze the Dragonfly source code. """
    import pep8
    argv = [
            sys.argv[0],
            "--filename=*.py",
           ]
    from pkg_resources import resource_filename
    setup_path = os.path.abspath(resource_filename(__name__, "setup.py"))
    directory = os.path.dirname(setup_path)
    pep8.process_options(argv)
    pep8.input_dir(directory)


#===========================================================================
# This script's main control logic.

def main():
    from dragonfly.log import setup_tracing
    setup_tracing(sys.stdout, limit=30)
#    run_pep8()
#    run_pylint()
    run_nose()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# encoding: utf-8

"""
Script to test the Key and Text action on X11 using the "xev" program.

There should be two lines printed for each key press: a down event and
an up event. Each line will contain the key symbol (e.g. Return).
There will be a small delay per key press for both test actions to help view
the output from "xev" as the keys are pressed.

X11Errors may be logged for keys remapped or not mapped by xkb.
"""

from __future__ import print_function

import subprocess
import sys
import threading
import time
import logging

from dragonfly import Key, Text

logging.basicConfig()

XEV_PATH = "/usr/bin/xev"


def main():
    # Define keys to type.
    keys = []
    alphas = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
    digits = "0123456789"
    keys.extend(alphas)
    keys.extend(digits)

    # Add common symbol characters. Use long names for reserved characters.
    symbols = "!@#%^&*()_+`~[]{}|\\;'\".<>?="
    keys.extend(symbols)
    keys.extend(["comma", "colon", "minus", "slash", "space"])

    # Add virtual keys. Some keys are repeated purposefully to maintain key
    # state (e.g. caps lock).
    keys.extend([
        # Whitespace and editing keys
        "enter", "tab", "space", "backspace", "delete",

        # Special keys
        "escape", "insert", "insert", "pause", "apps",

        # Lock keys
        "scrolllock", "scrolllock", "numlock", "numlock", "capslock",
        "capslock",

        # Navigation keys
        "up", "down", "left", "right", "pageup", "pagedown", "home", "end",

        # Number pad keys
        "npmul", "npadd", "npsub", "npdec", "npdiv", "npsep", "np0", "np1",
        "np2", "np3", "np4", "np5", "np6", "np7", "np8", "np9",

        # Function keys
        "f1", "f2", "f3", "f4", "f5", "f6", "f7", "f8", "f9", "f10", "f11",
        "f11", "f12", "f13", "f14", "f15", "f16", "f17", "f18", "f19",
        "f20", "f21", "f22", "f23", "f24",

        # Multimedia keys (press toggle keys twice)
        "volumeup", "volup", "volumedown", "voldown", "volumemute",
        "volmute", "tracknext", "trackprev", "playpause", "playpause",
        "browserback", "browserforward",

        # Include some keys not defined in typeables.py. There are a lot of
        # these in X11. The first two are browser-specific media keys. The
        # others are the Cyrillic л and Japanese ぁ letters.
        "XF86Refresh", "XF86HomePage", "Cyrillic_el", "U3041",

        # Modifiers.
        "shift:down", "shift:up",
        "ctrl:down", "ctrl:up",
        "alt:down", "alt:up",
        "rshift:down", "rshift:up",
        "rctrl:down", "rctrl:up",
        "ralt:down", "ralt:up",
        "win:down", "win:up", "win:down", "win:up",
        "rwin:down", "rwin:up", "rwin:down", "rwin:up",

        # Some common key combinations.
        "s-insert",
        "c-left",
        "c-home",
        "a-f",
        "c-a",
    ])

    # Define text to type using all alphanumeric characters and valid
    # symbols. Also test some Unicode characters.
    other_symbols = u"é—…ôêěèźżėфй№😱ß°⌷⅕€¨§™ぁ"
    text = alphas + digits + symbols + ",:-/ \n\t" + other_symbols

    # Start xev and selectively print lines from it in the background.
    # Exit if the command fails.
    try:
        proc = subprocess.Popen(XEV_PATH,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                stdin=subprocess.PIPE)
        def print_key_events():
            for line in iter(proc.stdout.readline, b''):
                line = line.decode('utf-8')
                if "keysym" in line:
                    print(line.strip())

        t = threading.Thread(target=print_key_events)
        t.setDaemon(True)  # make this thread terminate with the main thread
        t.start()
    except Exception as e:
        print("Couldn't start xev: %s" % e, file=sys.stderr)
        exit(1)

    # Wait a few seconds before beginning.
    print("Please ensure xev is focused...")
    time.sleep(5)

    print("--------------------Starting Key tests--------------------")
    keys.extend(other_symbols)
    Key("/10,".join(keys)).execute()
    print("Done.")

    print("--------------------Starting Text tests-------------------")
    Text(text, pause=0.1).execute()
    print("Done.")

    print("Stopping xev.")
    proc.terminate()


if __name__ == '__main__':
    main()



# -*- encoding: windows-1252 -*-
#
# This file is part of Dragonfly.
# (c) Copyright 2018 by Dane Finlay
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

import locale
import unittest

import six

from dragonfly.engines import EngineBase
from dragonfly import (Literal, Dictation, Sequence, CompoundRule,
                       get_engine)
from dragonfly.test import ElementTester, RecognitionFailure, RuleTestCase


# --------------------------------------------------------------------------

class TestEngineText(unittest.TestCase):

    def setUp(self):
        engine = get_engine("text")
        assert isinstance(engine, EngineBase)
        assert engine.name == "text"
        self.engine = engine

    def test_literal(self):
        """ Verify that the text engine is usable. """
        self.engine.speak("testing text")
        tester = ElementTester(Literal("hello world"))
        results = tester.recognize("hello world")
        assert results == "hello world"

        # Check that recognition failure is possible.
        results = tester.recognize("goodbye")
        assert results is RecognitionFailure

    def test_unicode_literals(self):
        """ Verify that the text engine can mimic literals using non-ascii
            characters. """
        tester = ElementTester(Literal(u"touché"))

        # Test that strings and Unicode objects can be used.
        string = "touché"
        if isinstance(string, six.binary_type):
            encoding = locale.getpreferredencoding()
            string = string.decode("windows-1252").encode(encoding)
        results = tester.recognize(string)
        assert results == u"touché"
        results = tester.recognize(u"touché")
        assert results == u"touché"

        # Check that recognition failure is possible.
        results = tester.recognize(u"jalapeño")
        assert results is RecognitionFailure



import unittest

from dragonfly.accessibility import utils


class AccessibilityTestCase(unittest.TestCase):

    def setUp(self):
        pass

    def assert_found_text(self, expected, query, expanded_text, cursor_offset=0):
        if cursor_offset == -1:
            cursor_offset = len(expanded_text)
        result = utils._find_text(query, expanded_text, cursor_offset)
        if expected is None:
            self.assertIsNone(result)
        else:
            self.assertIsNotNone(result)
            self.assertEqual(expected, expanded_text[result[0]:result[1]])

    def test_find_text(self):
        # Simple phrase selection.
        self.assert_found_text("elephant",
                               utils.TextQuery(end_phrase="elephant"),
                               "dog elephant tiger")
        self.assert_found_text("elephant ",
                               utils.TextQuery(end_phrase="elephant "),
                               "dog elephant tiger")
        self.assert_found_text("elephant. ",
                               utils.TextQuery(end_phrase="elephant. "),
                               "dog elephant. tiger")
        self.assert_found_text(" elephant",
                               utils.TextQuery(end_phrase=" elephant"),
                               "dog elephant tiger")
        self.assert_found_text("dog.",
                               utils.TextQuery(end_phrase="dog."),
                               "dog.elephant")
        self.assert_found_text("dog",
                               utils.TextQuery(end_phrase="dog", end_relative_position=utils.CursorPosition.BEFORE, end_relative_phrase="elephant"),
                               "dog elephant tiger")
        self.assert_found_text("elephant",
                               utils.TextQuery(end_phrase="elephant", end_relative_position=utils.CursorPosition.AFTER, end_relative_phrase="dog"),
                               "dog elephant tiger")
        self.assert_found_text(" ",
                               utils.TextQuery(end_phrase=" ", end_relative_position=utils.CursorPosition.BEFORE, end_relative_phrase="elephant"),
                               "dog elephant tiger")
        # Test matching against non-breaking space.
        self.assert_found_text(u"\u00a0",
                               utils.TextQuery(end_phrase=" ", end_relative_position=utils.CursorPosition.BEFORE, end_relative_phrase="elephant"),
                               u"dog\u00a0elephant tiger")
        self.assert_found_text(".word",
                               utils.TextQuery(end_phrase=".word"),
                               "..word..")
        self.assert_found_text("word.",
                               utils.TextQuery(end_phrase="word."),
                               "..word..")

        # Selecting a range.
        self.assert_found_text("dog elephant tiger",
                               utils.TextQuery(start_phrase="dog", through=True, end_phrase="tiger"),
                               "dog elephant tiger")
        self.assert_found_text("dog elephant tiger",
                               utils.TextQuery(start_phrase="dog", start_relative_position=utils.CursorPosition.BEFORE, start_relative_phrase="elephant", through=True, end_phrase="tiger"),
                               "dog elephant tiger")
        self.assert_found_text("sentence one.",
                               utils.TextQuery(start_phrase="sentence", through=True, end_phrase="."),
                               "sentence one. Sentence two.")
        self.assert_found_text(" one. Sentence two",
                               utils.TextQuery(start_phrase=" ", start_relative_position=utils.CursorPosition.BEFORE, start_relative_phrase="one", through=True, end_phrase="two"),
                               "sentence one. Sentence two.")
        self.assert_found_text(" elephant tiger",
                               utils.TextQuery(start_relative_position=utils.CursorPosition.AFTER, start_relative_phrase="dog", through=True, end_phrase="tiger"),
                               "doggy dog elephant tiger")
        self.assert_found_text(" elephant tiger",
                               utils.TextQuery(start_relative_position=utils.CursorPosition.AFTER, start_relative_phrase="doggy dog", through=True, end_phrase="tiger"),
                               "doggy dog elephant tiger")
        self.assert_found_text("dog elephant tigers ",
                               utils.TextQuery(start_phrase="dog", through=True, end_relative_position=utils.CursorPosition.BEFORE, end_relative_phrase="tiger"),
                               "dog elephant tigers tiger")
        # Test that we find the nearest overlapping match.
        self.assert_found_text("dog elephant",
                               utils.TextQuery(start_phrase="dog", through=True, end_phrase="elephant"),
                               "dog dog elephant",
                               cursor_offset=-1)

        # Selecting from the cursor position.
        self.assert_found_text("dog elephant",
                               utils.TextQuery(through=True, end_phrase="elephant"),
                               "dog elephant tiger",
                               cursor_offset=0)
        self.assert_found_text("elephant tiger",
                               utils.TextQuery(through=True, end_phrase="elephant"),
                               "dog elephant tiger",
                               cursor_offset=-1)
        self.assert_found_text("dog ",
                               utils.TextQuery(through=True, end_relative_position=utils.CursorPosition.BEFORE, end_relative_phrase="elephant"),
                               "dog elephant tiger",
                               cursor_offset=0)
        self.assert_found_text(" tiger",
                               utils.TextQuery(through=True, end_relative_position=utils.CursorPosition.AFTER, end_relative_phrase="elephant"),
                               "dog elephant tiger",
                               cursor_offset=-1)


if __name__ == "__main__":
    unittest.main()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Playback action
============================================================================

The :class:`Playback` action mimics a sequence of recognitions.  This is
for example useful for repeating a series of prerecorded or predefined
voice-commands.

This class could for example be used to reload with one single action::

    action = Playback([
                       (["focus", "Natlink"], 1.0),
                       (["File"], 0.5),
                       (["Reload"], 0.0),
                     ])
    action.execute()


Mimic quirks
----------------------------------------------------------------------------

Some SR engine back-ends have confusing :meth:`engine.mimic` method
behavior.  See the engine-specific mimic method documentation in sections
under :ref:`RefEngines` for more information.


Class reference
----------------------------------------------------------------------------

"""

from time              import sleep
from .action_base      import ActionBase, ActionError
from ..engines         import get_engine


#---------------------------------------------------------------------------

class Playback(ActionBase):
    """ Playback a series of recognitions. """

    def __init__(self, series, speed=1):
        """
            Constructor arguments:
             - *series* (sequence of 2-tuples) --
               the recognitions to playback.  Each element must be a
               2-tuple of the form *(["words", "two", "mimic"], interval)*,
               where *interval* is a float giving the number of seconds to
               pause after the given words are mimicked.
             - *speed* (*float*) --
               the factor by which to speed up playback.  The intervals
               after each mimic are divided by this number.

        """
        ActionBase.__init__(self)
        self._series = tuple(series)
        self._speed = float(speed)
        self._str = str([w for w, i in self._series])

    def _get_speed(self):
        return self._speed
    def _set_speed(self, speed):
        self._speed = float(speed)
    speed = property(fget=_get_speed, fset=_set_speed,
                     doc="Factor to speed up playback.")

    def _execute(self, data=None):
        engine = get_engine()

        # Mimic the series of recognitions.
        for words, interval in self._series:
            self._log.debug("Mimicking recognition: %r", words)
            try:
                engine.mimic(words)
                if interval and self._speed:
                    sleep(interval / self._speed)
            except Exception as e:
                raise ActionError("Playback failed: %s" % e)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
SAPI 5 engine classes
============================================================================

"""


#---------------------------------------------------------------------------

import logging
import time
import os.path

from datetime import datetime
from ctypes import Structure, c_long, c_int, c_uint, pointer
from six import string_types, integer_types

import pythoncom
import win32con
from win32com.client           import Dispatch, getevents, constants
from win32com.client.gencache  import EnsureDispatch
from ctypes import windll, WinError, WINFUNCTYPE
from ctypes.wintypes import DWORD, HANDLE, HWND, LONG

from ..base                    import (EngineBase, EngineError,
                                       MimicFailure, DelegateTimerManager,
                                       DelegateTimerManagerInterface,
                                       DictationContainerBase,
                                       GrammarWrapperBase)
from .compiler                 import Sapi5Compiler
from .recobs                   import Sapi5RecObsManager
from ...grammar.state          import State
from ...grammar.recobs         import RecognitionObserver
from ...windows.window         import Window


#===========================================================================

class POINT(Structure):
    _fields_ = [('x', c_long),
                ('y', c_long)]

class MSG(Structure):
    _fields_ = [('hwnd', c_int),
                ('message', c_uint),
                ('wParam', c_int),
                ('lParam', c_int),
                ('time', c_int),
                ('pt', POINT)]


class MimicObserver(RecognitionObserver):
    _log = logging.getLogger("SAPI5 RecObs")

    def __init__(self):
        RecognitionObserver.__init__(self)
        self.status = "none"

    def on_recognition(self, words):
        self._log.debug("SAPI5 RecObs on_recognition(): %r" % (words,))
        self.status = "recognition: %r" % (words,)

    def on_failure(self):
        self._log.debug("SAPI5 RecObs on_failure()")
        self.status = "failure"


#===========================================================================

class Sapi5SharedEngine(EngineBase, DelegateTimerManagerInterface):
    """ Speech recognition engine back-end for SAPI 5 shared recognizer. """

    _name = "sapi5shared"
    recognizer_dispatch_name = "SAPI.SpSharedRecognizer"
    DictationContainer = DictationContainerBase

    #-----------------------------------------------------------------------

    def __init__(self, retain_dir=None):
        """
        :param retain_dir: Retains recognized audio and/or  metadata in the
          given directory, saving audio to ``retain_[timestamp].wav`` file
          and metadata to ``retain.tsv``.

          Disabled by default (``None``).
        :type retain_dir: str|None
        """
        EngineBase.__init__(self)
        DelegateTimerManagerInterface.__init__(self)

        EnsureDispatch(self.recognizer_dispatch_name)
        EnsureDispatch("SAPI.SpVoice")
        self._recognizer  = None
        self._speaker     = None
        self._compiler    = None

        self._recognition_observer_manager = Sapi5RecObsManager(self)
        self._timer_manager = DelegateTimerManager(0.02, self)

        if isinstance(retain_dir, string_types) or retain_dir is None:
            self._retain_dir = retain_dir
        else:
            self._retain_dir = None
            self._log.error("Invalid retain_dir: %r" % retain_dir)

    def connect(self):
        """ Connect to back-end SR engine. """
        self._recognizer  = Dispatch(self.recognizer_dispatch_name)
        self._speaker     = Dispatch("SAPI.SpVoice")
        self._compiler    = Sapi5Compiler()

    def disconnect(self):
        """ Disconnect from back-end SR engine. """
        self._recognizer  = None
        self._speaker     = None
        self._compiler    = None

    #-----------------------------------------------------------------------
    # Methods for working with grammars.

    def _load_grammar(self, grammar):
        """ Load the given *grammar*. """
        self._log.debug("Loading grammar %s." % grammar.name)
        if not self._recognizer:
            self.connect()

        # Create recognition context, compile grammar, and create
        #  the grammar wrapper object for managing this grammar.
        context = self._recognizer.CreateRecoContext()
        if self._retain_dir:
            context.RetainedAudio = constants.SRAORetainAudio
        handle = self._compiler.compile_grammar(grammar, context)
        wrapper = GrammarWrapper(grammar, handle, context, self,
                                 self._recognition_observer_manager)

        handle.State = constants.SGSEnabled
        for rule in grammar.rules:
            handle.CmdSetRuleState(rule.name, constants.SGDSActive)
       # self.activate_grammar(grammar)
       # for l in grammar.lists:
       #     l._update()
        handle.CmdSetRuleState("_FakeRule", constants.SGDSActive)

        return wrapper

    def _unload_grammar(self, grammar, wrapper):
        """ Unload the given *grammar*. """
        try:
            wrapper.handle.State = constants.SGSDisabled
        except Exception as e:
            self._log.exception("Failed to unload grammar %s: %s."
                                % (grammar, e))

    def activate_grammar(self, grammar):
        """ Activate the given *grammar*. """
        self._log.debug("Activating grammar %s." % grammar.name)
        grammar_handle = self._get_grammar_wrapper(grammar).handle
        grammar_handle.State = constants.SGSEnabled

    def deactivate_grammar(self, grammar):
        """ Deactivate the given *grammar*. """
        self._log.debug("Deactivating grammar %s." % grammar.name)
        grammar_handle = self._get_grammar_wrapper(grammar).handle
        grammar_handle.State = constants.SGSDisabled

    def activate_rule(self, rule, grammar):
        """ Activate the given *rule*. """
        self._log.debug("Activating rule %s in grammar %s."
                        % (rule.name, grammar.name))
        grammar_handle = self._get_grammar_wrapper(grammar).handle
        grammar_handle.CmdSetRuleState(rule.name, constants.SGDSActive)

    def deactivate_rule(self, rule, grammar):
        """ Deactivate the given *rule*. """
        self._log.debug("Deactivating rule %s in grammar %s."
                        % (rule.name, grammar.name))
        grammar_handle = self._get_grammar_wrapper(grammar).handle
        grammar_handle.CmdSetRuleState(rule.name, constants.SGDSInactive)

    def update_list(self, lst, grammar):
        grammar_handle = self._get_grammar_wrapper(grammar).handle
        list_rule_name = "__list_%s" % lst.name
        rule_handle = grammar_handle.Rules.FindRule(list_rule_name)

        rule_handle.Clear()
        src_state = rule_handle.InitialState
        dst_state = None
        for item in lst.get_list_items():
            src_state.AddWordTransition(dst_state, item)

        grammar_handle.Rules.Commit()

    def set_exclusiveness(self, grammar, exclusive):
        self._log.debug("Setting exclusiveness of grammar %s to %s."
                        % (grammar.name, exclusive))
        wrapper = self._get_grammar_wrapper(grammar)
        if exclusive and wrapper.handle.State != constants.SGSExclusive:
            wrapper.state_before_exclusive = wrapper.handle.State
            wrapper.handle.State = constants.SGSExclusive
        elif not exclusive and wrapper.handle.State == constants.SGSExclusive:
            assert wrapper.state_before_exclusive in (constants.SGSEnabled,
                                                      constants.SGSDisabled)
            wrapper.handle.State = wrapper.state_before_exclusive
        # grammar_handle.SetGrammarState(constants.SPGS_EXCLUSIVE)


    #-----------------------------------------------------------------------
    # Miscellaneous methods.

    def mimic(self, words):
        """
        Mimic a recognition of the given *words*.

        .. note:: This method has a few quirks to be aware of:

           #. Mimic can fail to recognize a command if the relevant grammar
              is not yet active.
           #. Mimic does not work reliably with the shared recognizer unless
              there are one or more exclusive grammars active.
           #. Mimic can **crash the process** in some circumstances, e.g.
              when mimicking non-ASCII characters.

        """
        self._log.debug("SAPI5 mimic: %r" % (words,))
        if isinstance(words, string_types):
            phrase = words
        else:
            phrase = " ".join(words)

        # Fail on empty input.
        if not phrase:
            raise MimicFailure("Invalid mimic input %r" % phrase)

        # Register a recognition observer for checking the success of this
        # mimic.
        observer = MimicObserver()
        observer.register()

        # Emulate recognition of the phrase and wait for recognition to
        # finish, timing out after 2 seconds.
        self._recognizer.EmulateRecognition(phrase)
        timeout = 2
        NULL = c_int(win32con.NULL)
        if timeout != None:
            begin_time = time.time()
            windll.user32.SetTimer(NULL, NULL, int(timeout * 1000), NULL)

        message = MSG()
        message_pointer = pointer(message)
        while (not timeout) or (time.time() - begin_time < timeout):
            if timeout:
                self._log.debug("SAPI5 message loop: %s sec left"
                                % (timeout + begin_time - time.time()))
            else:
                self._log.debug("SAPI5 message loop: no timeout")

            if windll.user32.GetMessageW(message_pointer, NULL, 0, 0) == 0:
                msg = str(WinError())
                self._log.error("GetMessageW() failed: %s" % msg)
                raise EngineError("GetMessageW() failed: %s" % msg)

            self._log.debug("SAPI5 message: %r" % (message.message,))
            if message.message == win32con.WM_TIMER:
                # A timer message means this loop has timed out.
                self._log.debug("SAPI5 message loop timed out: %s sec left"
                                % (timeout + begin_time - time.time()))
                break
            else:
                # Process other messages as normal.
                self._log.debug("SAPI5 message translating and dispatching.")
                windll.user32.TranslateMessage(message_pointer)
                windll.user32.DispatchMessageW(message_pointer)

            if observer.status.startswith("recognition:"):
                # The previous message was a recognition which matched.
                self._log.debug("SAPI5 message caused recognition.")

        # Unregister the observer and check its status.
        observer.unregister()
        if observer.status == "failure":
            raise MimicFailure("Mimic failed.")
        elif observer.status == "none":
            raise MimicFailure("Mimic failed, nothing happened.")

    def speak(self, text):
        """ Speak the given *text* using text-to-speech. """
        self._speaker.Speak(text)

    def _get_language(self):
        if not self._recognizer:
            return "en"

        # Get Windows language identifiers for supported languages from the
        # recognizer's current status information.
        languages = self._recognizer.Status.SupportedLanguages

        # Lookup and return the language tag for the first supported
        # language ID.
        if languages:
            return self._get_language_tag(languages[0])
        else:
            return "en"

    def _has_quoted_words_support(self):
        return False

    def _do_recognition(self):
        """
            Recognize speech in a loop.

            This will also call any scheduled timer functions and ensure
            that the correct window context is used.

        """

        # Register for window change events to activate/deactivate grammars
        # and rules on window changes, including window title changes. This
        # is done here because the SAPI5 'OnPhraseStart' grammar callback is
        # called after grammar state changes are allowed.
        WinEventProcType = WINFUNCTYPE(None, HANDLE, DWORD, HWND, LONG,
                                       LONG, DWORD, DWORD)

        self._last_foreground_window = None
        self._last_foreground_window_title = None

        def callback(hWinEventHook, event, hwnd, idObject, idChild,
                     dwEventThread, dwmsEventTime):
            window = Window.get_foreground()
            # Note: hwnd doesn't always match window.handle, even when
            # foreground window changed (and sometimes it didn't change)
            window_changed = (
                window != self._last_foreground_window or
                window == self._last_foreground_window and
                window.title != self._last_foreground_window_title
            )
            if window_changed:
                self.process_grammars_context(window)
                self._last_foreground_window = window
                self._last_foreground_window_title = window.title

        def set_hook(win_event_proc, event_type):
            return windll.user32.SetWinEventHook(
                event_type, event_type, 0, win_event_proc, 0, 0,
                win32con.WINEVENT_OUTOFCONTEXT)

        win_event_proc = WinEventProcType(callback)
        windll.user32.SetWinEventHook.restype = HANDLE

        [set_hook(win_event_proc, et) for et in
         {win32con.EVENT_SYSTEM_FOREGROUND,
          win32con.EVENT_OBJECT_NAMECHANGE, }]

        # Recognize speech, call timer functions and handle window change
        # events in a loop. Stop on disconnect().
        self.speak('beginning loop!')
        while self._recognizer is not None:
            pythoncom.PumpWaitingMessages()
            self.call_timer_callback()
            time.sleep(0.005)

#---------------------------------------------------------------------------
# Make the shared engine available as Sapi5Engine, for backwards
#  compatibility.

Sapi5Engine = Sapi5SharedEngine


#===========================================================================

class Sapi5InProcEngine(Sapi5SharedEngine):
    """
        Speech recognition engine back-end for SAPI 5 in process
        recognizer.

    """

    _name = "sapi5inproc"
    recognizer_dispatch_name = "SAPI.SpInProcRecognizer"

    def connect(self, audio_source=0):
        """
            Connect to the speech recognition backend.

            The audio source to use for speech recognition can be
            specified using the *audio_source* argument. If it is not
            given, it defaults to the first audio source found.

        """

        Sapi5SharedEngine.connect(self)
        self.select_audio_source(audio_source)

    def get_audio_sources(self):
        """
            Get the available audio sources.

            This method returns a list of audio sources, each represented
            by a 3-element tuple: the index, the description, and the COM
            handle for the audio source.

        """

        available_sources = self._recognizer.GetAudioInputs()
        audio_sources_list = []
        for index, item in enumerate(collection_iter(available_sources)):
            audio_sources_list.append((index, item.GetDescription(), item))
        return audio_sources_list

    def select_audio_source(self, audio_source):
        """
            Configure the speech recognition engine to use the given
            audio source.

            The audio source may be specified as follows:
             - As an *int* specifying the index of the audio source to use
             - As a *str* containing the description of the audio source
               to use, or a substring thereof

            The :meth:`get_audio_sources()` method can be used to
            retrieve the available sources together with their indices
            and descriptions.

        """

        available_sources = self._recognizer.GetAudioInputs()

        if isinstance(audio_source, integer_types):
            # Parameter is the index of the source to use.
            if 0 <= audio_source < available_sources.Count:
                selected_source = available_sources.Item(audio_source)
            else:
                raise EngineError("Invalid audio source index: %r"
                                  " (%s sources available, so index must be"
                                  " in range 0 to %s)"
                                  % (audio_source, available_sources.Count,
                                     available_sources.Count - 1))

        elif isinstance(audio_source, string_types):
            for item in collection_iter(available_sources):
                if audio_source in item.GetDescription():
                    selected_source = item
                    break
            else:
                raise EngineError("Audio source not found: %r"
                                  % (audio_source))

        else:
            raise EngineError("Invalid audio source qualifier: %r"
                              % (audio_source))

        self._log.info("Selecting audio source: %r"
                       % (selected_source.GetDescription(),))
        self._recognizer.AudioInput = selected_source


#---------------------------------------------------------------------------
# Utility generator function for iterating over COM collections.

def collection_iter(collection):
    if not collection:
        return
    for index in range(0, collection.Count):
        yield collection.Item(index)


#---------------------------------------------------------------------------

class GrammarWrapper(GrammarWrapperBase):

    def __init__(self, grammar, handle, context, engine, recobs_manager):
        GrammarWrapperBase.__init__(self, grammar, engine, recobs_manager)
        self.handle = handle
        self.context = context
        self.state_before_exclusive = handle.State

        # Register callback functions which will handle recognizer events.
        base = getevents("SAPI.SpSharedRecoContext")
        class ContextEvents(base): pass
        c = ContextEvents(context)
        c.OnPhraseStart = self.phrase_start_callback
        c.OnRecognition = self.recognition_callback
        if hasattr(grammar, "process_recognition_other"):
            c.OnRecognitionForOtherContext = self.recognition_other_callback
        if hasattr(grammar, "process_recognition_failure"):
            c.OnFalseRecognition = self.recognition_failure_callback

    def phrase_start_callback(self, stream_number, stream_position):
        window = Window.get_foreground()
        self.grammar.process_begin(window.executable, window.title,
                                   window.handle)

    def recognition_callback(self, StreamNumber, StreamPosition,
                             RecognitionType, Result):
        try:
            newResult = Dispatch(Result)
            phrase_info = newResult.PhraseInfo
            rule_name = phrase_info.Rule.Name

            #---------------------------------------------------------------
            # Build a list of rule names for each element.

            # First populate it with the top level rule name.
            element = phrase_info.Rule
            name = element.Name
            start = element.FirstElement
            count = element.NumberOfElements
            rule_names = [name] * count

            # Walk the tree of child rules and put their names in the list.
            stack = [collection_iter(phrase_info.Rule.Children)]
            while stack:
                try: element = next(stack[-1])
                except StopIteration: stack.pop(); continue
                name = element.Name
                start = element.FirstElement
                count = element.NumberOfElements
                rule_names[start:start + count] = [name] * count
                if element.Children:
                    stack.append(collection_iter(element.Children))

            #---------------------------------------------------------------
            # Prepare the words and rule names for the element parsers.

            replacements = [False] * len(rule_names)
            if phrase_info.Replacements:
                for replacement in collection_iter(phrase_info.Replacements):
                    begin = replacement.FirstElement
                    end = begin + replacement.NumberOfElements
                    replacements[begin] = replacement.Text
                    for index in range(begin + 1, end):
                        replacements[index] = True

            results = []
            rule_set = list(set(rule_names))

            elements = phrase_info.Elements
            for index in range(len(rule_names)):
                element = elements.Item(index)
                rule_id = rule_set.index(rule_names[index])

                # Map dictation rule IDs to 1M so that dragonfly recognizes
                # the words as dictation.
                if rule_names[index] == "dgndictation":
                    rule_id = 1000000

                replacement = replacements[index]
                info = [element.LexicalForm, rule_id,
                        element.DisplayText, element.DisplayAttributes,
                        replacement]
                results.append(info)

            #---------------------------------------------------------------
            # Retain audio

            # Only write audio data and metadata if the directory exists.
            retain_dir = self.engine._retain_dir
            if retain_dir and not os.path.isdir(retain_dir):
                self.engine._log.warning(
                    "Audio was not retained because '%s' was not a "
                    "directory" % retain_dir
                )
            elif retain_dir:
                try:
                    file_stream = Dispatch("SAPI.SpFileStream")
                    # Note: application can also retrieve smaller portions
                    # of the audio stream by specifying a starting phrase
                    # element and phrase element length.
                    audio_stream = newResult.Audio()

                    # Make sure we have audio data, which we wouldn't from a
                    # mimic or if the retain flag wasn't set above.
                    if audio_stream:
                        # Write audio data.
                        file_stream.Format = audio_stream.Format
                        now = datetime.now()
                        filename = ("retain_%s.wav"
                                    % now.strftime("%Y-%m-%d_%H-%M-%S_%f"))
                        wav_path = os.path.join(retain_dir, filename)
                        flags = constants.SSFMCreateForWrite
                        file_stream.Open(wav_path, flags)
                        try:
                            file_stream.Write(audio_stream.GetData())
                        finally:
                            file_stream.Close()

                        # Write metadata
                        words = ' '.join([r[2] for r in results])
                        audio_length = int(newResult.Times.Length) / 1e7
                        tsv_path = os.path.join(retain_dir, "retain.tsv")
                        with open(tsv_path, "a") as tsv_file:
                            tsv_file.write('\t'.join([
                                filename, str(audio_length),
                                self.grammar.name, rule_name, words
                            ]) + '\n')
                except:
                    self.engine._log.exception("Exception retaining audio")

            #---------------------------------------------------------------
            # Attempt to parse the recognition.

            func = getattr(self.grammar, "process_recognition", None)
            words = tuple([r[0] for r in results])
            if func:
                if not self._process_grammar_callback(func, words=words,
                                                      results=newResult):
                    return

            s = State(results, rule_set, self.engine)
            for r in self.grammar.rules:
                if not (r.active and r.exported):
                    continue

                s.initialize_decoding()
                for result in r.decode(s):
                    if s.finished():
                        # Notify recognition observers, then process the
                        # rule.
                        root = s.build_parse_tree()
                        notify_args = (words, r, root, newResult)
                        self.recobs_manager.notify_recognition(*notify_args)
                        r.process_recognition(root)
                        self.recobs_manager.notify_post_recognition(
                            *notify_args
                        )
                        return

        except Exception as e:
            Sapi5Engine._log.error("Grammar %s: exception: %s"
                                   % (self.grammar._name, e), exc_info=True)

        #-------------------------------------------------------------------
        # If this point is reached, then the recognition was not
        #  processed successfully..

        self.engine._log.error("Grammar %s: failed to decode"
                               " recognition %r."
                               % (self.grammar._name,
                                  [r[0] for r in results]))

    def recognition_other_callback(self, StreamNumber, StreamPosition):
        # Note that SAPI 5.3 doesn't offer access to the actual
        #  recognition contents during a
        #  OnRecognitionForOtherContext event.
        func = getattr(self.grammar, "process_recognition_other", None)
        self._process_grammar_callback(func, words=False, results=None)

    def recognition_failure_callback(self, StreamNumber, StreamPosition,
                                     Result):
        func = getattr(self.grammar, "process_recognition_failure", None)
        self._process_grammar_callback(func, results=Dispatch(Result))



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Recognition state change callbacks
----------------------------------------------------------------------------

"""

from .recobs import RecognitionObserver

try:
    from inspect import getfullargspec as getargspec
except ImportError:
    # Fallback on the deprecated function.
    from inspect import getargspec


class CallbackRecognitionObserver(RecognitionObserver):
    """
    Observer class for calling recognition state change callbacks.

    This class is used by the ``register_*_callback`` functions and is
    registered with the current engine on initialization.

    Constructor arguments:
     - *event* (*str*) -- the name of the recognition event to register for
       (e.g. ``"on_begin"``).
     - *function* (*callable*) -- function to call on the specified
       recognition event.

    """
    def __init__(self, event, function):
        RecognitionObserver.__init__(self)
        if not hasattr(self, event):
            raise ValueError("'%s' is not a valid recognition event"
                             % event)
        if not callable(function):
            raise TypeError("callback function for event '%s' is not "
                            "callable" % event)
        self._function = function
        self._event = event
        self.register()

    def _process_recognition_event(self, cb_name, required_names, **kwargs):
        if not (self._event == cb_name and callable(self._function)):
            return

        # If the callback function takes keyword arguments, only send those
        # that it accepts. Always pass required names.
        func_kwargs = kwargs
        if func_kwargs:
            argspec = getargspec(self._function)
            arg_names, kwargs_names = argspec[0], argspec[2]
            if not kwargs_names:
                func_kwargs = {k: v for (k, v) in func_kwargs.items()
                               if k in arg_names or k in required_names}

        # Call the callback function.
        self._function(**func_kwargs)

    def on_begin(self):
        """"""
        self._process_recognition_event("on_begin", [])

    def on_recognition(self, words, rule, node, results):
        """"""
        self._process_recognition_event("on_recognition", ["words"],
                                        words=words, rule=rule, node=node,
                                        results=results)

    def on_failure(self, results):
        """"""
        self._process_recognition_event("on_failure", [], results=results)

    def on_end(self, results):
        """"""
        self._process_recognition_event("on_end", [], results=results)

    def on_post_recognition(self, words, rule, node, results):
        """"""
        self._process_recognition_event("on_post_recognition", ["words"],
                                        words=words, rule=rule, node=node,
                                        results=results)


def register_beginning_callback(function):
    """
    Register a callback function to be called when speech starts.

    The :class:`CallbackRecognitionObserver` object returned from this
    function can be used to unregister the callback function.

    :param function: callback function
    :type function: callable
    :returns: recognition observer
    :rtype: CallbackRecognitionObserver
    """
    return CallbackRecognitionObserver("on_begin", function)


def register_recognition_callback(function):
    """
    Register a callback function to be called on recognition success.

    The :class:`CallbackRecognitionObserver` object returned from this
    function can be used to unregister the callback function.

    :param function: callback function
    :type function: callable
    :returns: recognition observer
    :rtype: CallbackRecognitionObserver
    """
    return CallbackRecognitionObserver("on_recognition", function)


def register_failure_callback(function):
    """
    Register a callback function to be called on recognition failures.

    The :class:`CallbackRecognitionObserver` object returned from this
    function can be used to unregister the callback function.

    :param function: callback function
    :type function: callable
    :returns: recognition observer
    :rtype: CallbackRecognitionObserver
    """
    return CallbackRecognitionObserver("on_failure", function)


def register_ending_callback(function):
    """
    Register a callback function to be called when speech ends, either
    successfully (after calling the recognition callback) or in failure
    (after calling the failure callback).

    The :class:`CallbackRecognitionObserver` object returned from this
    function can be used to unregister the callback function.

    :param function: callback function
    :type function: callable
    :returns: recognition observer
    :rtype: CallbackRecognitionObserver
    """
    return CallbackRecognitionObserver("on_end", function)


def register_post_recognition_callback(function):
    """
    Register a callback function to be called after all rule processing
    has completed after recognition success.

    The :class:`CallbackRecognitionObserver` object returned from this
    function can be used to unregister the callback function.

    :param function: callback function
    :type function: callable
    :returns: recognition observer
    :rtype: CallbackRecognitionObserver
    """
    return CallbackRecognitionObserver("on_post_recognition", function)



# -*- encoding: windows-1252 -*-
#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

import copy
import locale
import unittest

from six import string_types, text_type, binary_type

from dragonfly.engines.base.dictation   import DictationContainerBase
from dragonfly.engines                  import get_engine
from dragonfly.grammar.elements         import Compound, Dictation
from dragonfly.test.element_testcase    import ElementTestCase
from dragonfly.test.element_tester      import ElementTester


#===========================================================================
class DictationElementTestCase(ElementTestCase):

    input_output = []

    def test_element(self):
        # Translate all inputs and outputs to the preferred encoding for
        # recognition so that the tests can run on Windows or Linux.
        engine = get_engine()
        encoding = locale.getpreferredencoding()
        uppercase_dictation_required = engine.name in ['sphinx', 'text']

        def translate(string):
            # Any binary strings in this file will be encoded with
            # windows-1252 as per the specified source file encoding.
            if isinstance(string, binary_type):
                string = string.decode("windows-1252").encode(encoding)

            return string

        for i, (input, output) in enumerate(self.input_output):
            new_input = translate(input)
            new_output = translate(output)

            # Also map input dictation words to uppercase if necessary.
            if uppercase_dictation_required:
                words = new_input.split()
                new_input = ' '.join(
                    [words[0]] + [word.upper() for word in words[1:]]
                )

                # Manually uppercase the accented characters we use because
                # upper() doesn't do it for us. There's probably a better
                # way, but this will do.
                if isinstance(new_input, binary_type):
                    new_input = (new_input.replace("\xa9", "\x89")
                                 .replace("\xb1", "\x91"))

            self.input_output[i] = (new_input, new_output)

        ElementTestCase.test_element(self)


class TestNonAsciiDictation(unittest.TestCase):

    def test_dictation_non_ascii(self):
        """ Test handling of non-ASCII characters in dictation. """

        def value_func(node, extras):
            return extras["text"]
        element = Compound("test <text>",
                           extras=[Dictation(name="text")],
                           value_func=value_func)
        tester = ElementTester(element)
        words = [u"test", u"touché"]
        engine = get_engine()
        uppercase_dictation_required = engine.name in ['sphinx', 'text']
        if uppercase_dictation_required:
            words[1] = words[1].upper()
        dictation = tester.recognize(words)

        # Verify recognition returned dictation result.
        if not isinstance(dictation, DictationContainerBase):
            encoding = locale.getpreferredencoding()
            message = (u"Expected recognition result to be a dictation"
                       u" container, but received %r"
                       % (repr(dictation),))
            self.fail(message.encode(encoding))

        # Verifying dictation converts/encode successfully.
        string = "touché"
        if isinstance(string, binary_type):
            encoding = locale.getpreferredencoding()
            string = string.decode("windows-1252").encode(encoding)
        self.assertEqual(str(dictation), string)
        self.assertEqual(text_type(dictation), u"touché")
        self.assertTrue(isinstance(repr(dictation), string_types))


class NonAsciiUnicodeDictationTestCase(DictationElementTestCase):
    """ Verify handling of non-ASCII characters in unicode dictation. """

    def _build_element(self):
        def value_func(node, extras):
            return text_type(extras["text"])
        return Compound("test <text>",
                        extras=[Dictation(name="text")],
                        value_func=value_func)

    input_output = [
                    (u"test dictation",     u"dictation"),
                    (u"test touché",        u"touché"),
                    (u"test jalapeño",      u"jalapeño"),
                   ]


class NonAsciiStrDictationTestCase(DictationElementTestCase):
    """ Verify handling of non-ASCII characters in str dictation. """

    def _build_element(self):
        def value_func(node, extras):
            return str(extras["text"])
        return Compound("test <text>",
                        extras=[Dictation(name="text")],
                        value_func=value_func)

    input_output = [
                    ("test dictation",     "dictation"),
                    ("test touché",        "touché"),
                    ("test jalapeño",      "jalapeño"),
                   ]


class FormattedDictationTestCase(DictationElementTestCase):
    """ Verify handling of string methods applied to Dictation objects """

    def _build_element(self):
        def value_func(node, extras):
            return str(extras["text"])
        return Compound("test <text>",
                        extras=[Dictation(name="text").upper().replace(" ", "/")],
                        value_func=value_func)

    input_output = [
                    ("test some random dictation", "SOME/RANDOM/DICTATION"),
                    ("test touché jalapeño",       "TOUCHÉ/JALAPEÑO"),
                   ]


class CamelDictationTestCase(DictationElementTestCase):
    """ Verify handling of camelCase formatting applied to Dictation objects """

    def _build_element(self):
        def value_func(node, extras):
            return str(extras["text"])
        return Compound("test <text>",
                        extras=[Dictation(name="text").camel()],
                        value_func=value_func)

    input_output = [
                    ("test some random dictation", "someRandomDictation"),
                    ("test touché jalapeño",       "touchéJalapeño"),
                   ]


class ApplyDictationTestCase(DictationElementTestCase):
    """ Verify handling of arbitrary formatting applied to Dictation objects using apply() """

    f = lambda self, s: "".join(
        "_".join(s.split(" ")[:-1] + [s.split(" ")[-1].upper()])
    )

    def _build_element(self):
        def value_func(node, extras):
            return str(extras["text"])
        return Compound("test <text>",
                        extras=[Dictation(name="text").apply(self.f)],
                        value_func=value_func)

    input_output = [
                    ("test some random dictation", "some_random_DICTATION"),
                    ("test touché jalapeño",       "touché_JALAPEÑO"),
                   ]


class DictationCopyTestCase(unittest.TestCase):

    def test_copy(self):
        """ Test that Dictation elements can be copied. """
        element = Dictation("text")
        self.assertIsNot(element, copy.copy(element))
        element = Dictation("text").camel()
        self.assertIsNot(element, copy.copy(element))

#===========================================================================

if __name__ == "__main__":
    unittest.main()




'''
ShortIntegerRef
============================================================================

:class:`ShortIntegerRef` is a modified version of :class:`IntegerRef` which allows for greater flexibility in the way that numbers may be pronounced, allowing for words like "hundred" to be dropped. This may be particularly useful when navigating files by line or page number.

Some examples of allowed pronunciations:

================================     ======
Pronunciation                        Result
================================     ======
one                                   1
ten                                   10
twenty three                          23
two three                             23
seventy                               70
seven zero                            70
hundred                               100
one oh three                          103
hundred three                         103
one twenty seven                      127
one two seven                         127
one hundred twenty seven              127
seven hundred                         700
thousand                              1000
seventeen hundred                     1700
seventeen hundred fifty three         1753
seventeen fifty three                 1753
one seven five three                  1753
seventeen five three                  1753
four thousand                         4000
================================     ======

The class works in the same way as :class:`IntegerRef`, by adding the following as an extra.

.. code:: python

   ShortIntegerRef("name", 0, 1000),
'''

from ..base.integer_internal  import (MapIntBuilder, CollectionIntBuilder,
                                      MagnitudeIntBuilder, IntegerContentBase)
from .number import int_0, int_1_9, int_10_19, int_and_1_99, int_20_90_10, int_100s, int_1000000s

# Twenty five
int_20_99       = MagnitudeIntBuilder(
                   factor      = 10,
                   spec        = "<multiplier> [<remainder>]",
                   multipliers = [int_20_90_10],
                   remainders  = [int_1_9],
                  )
# Two five / seven zero
int_10_99       = MagnitudeIntBuilder(
                   factor      = 10,
                   spec        = "<multiplier> <remainder>",
                   multipliers = [int_1_9],
                   remainders  = [int_0, int_1_9],
                  )
# Oh five
int_and_1_9    = CollectionIntBuilder(
                   spec        = "(oh | zero) <element>",
                   set         = [int_1_9],
                  )
# Fifty five / five five
int_and_10_99    = CollectionIntBuilder(
                   spec        = "[and] <element>",
                   set         = [int_10_19, int_20_99, int_10_99],
                  )
# Hundred fifty / two hundred / four hundred seventy nine
int_x01_x99        = MagnitudeIntBuilder(
                   factor      = 100,
                   spec        = "[<multiplier>] hundred [<remainder>]",
                   multipliers = [int_1_9, int_10_19, int_20_99],
                   remainders  = [int_1_9, int_and_10_99],
                  )
# One oh nine / five fifty / one two five
int_x10_x99        = MagnitudeIntBuilder(
                   factor      = 100,
                   spec        = "<multiplier> [hundred] <remainder>",
                   multipliers = [int_1_9, int_10_19, int_20_99],
                   remainders  = [int_and_1_9, int_and_10_99],
                  )
# One thousand fifty
int_x000_x099       = MagnitudeIntBuilder(
                   factor      = 1000,
                   spec        = "[<multiplier>] thousand [<remainder>]",
                   multipliers = [int_1_9],
                   remainders  = [int_1_9, int_and_10_99]
                  )
# One thousand five fifty / one five fifty / one five five zero
int_x100_x999       = MagnitudeIntBuilder(
                   factor      = 1000,
                   spec        = "<multiplier> [thousand] <remainder>",
                   multipliers = [int_1_9],
                   remainders  = [int_x01_x99, int_x10_x99]
                  )

# Very similar to the one from number.py, but without int_1_9 multipliers
# as these are handled above.
int_1000s       = MagnitudeIntBuilder(
                   factor      = 1000,
                   spec        = "[<multiplier>] thousand [<remainder>]",
                   multipliers = [int_10_19, int_20_99, int_100s],
                   remainders  = [int_and_1_99, int_100s]
                  )
#---------------------------------------------------------------------------

class ShortIntegerContent(IntegerContentBase):
    builders = [int_0, int_1_9, int_10_19, int_20_99, int_10_99,
                int_x01_x99, int_x10_x99, int_x000_x099, int_x100_x999,
                int_1000s, int_1000000s]



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
BasicRule class
============================================================================

The BasicRule class is designed to make it easy to create a rule from an
element tree, rather than building one indirectly via compound element
specs.

This rule class has the following parameters to customize its behavior:

 - *name* (*str*) -- the rule's name
 - *element* (*Element*) -- root element for this rule
 - *exported* -- whether the rule is exported
 - *context* -- context in which the rule will be active

Each of these parameters can be passed as a (keyword) arguments to the
constructor, or defined as a class attribute in a derived class.

.. note::

   The BasicRule class has only limited support for the "extras" and
   "defaults" functionality of the
   :class:`dragonfly.grammar.rule_compound.CompoundRule` and
   :class:`dragonfly.grammar.rule_mapping.MappingRule` classes. By default,
   the *extras* dictionary passed to :meth:`_process_recognition` will
   only contain an entry for the root element of the rule.


Example usage
............................................................................

The BasicRule class can be used to define a voice-command as follows::

   from dragonfly import (BasicRule, Repetition, Alternative, Literal, Text,
                          Grammar)

   class ExampleRule(BasicRule):
       # Define a rule element that accepts 1 to 5 (exclusive) repetitions
       # of either 'test one', 'test two' or 'test three'. These commands
       # type their respective numbers in succession using the Text action.
       element = Repetition(
           Alternative((
               Literal("test one", value=Text("1")),
               Literal("test two", value=Text("2")),
               Literal("test three", value=Text("3")),
           )),
           1, 5
       )

   # Create a grammar with the example rule and load it.
   rule = ExampleRule()
   grammar = Grammar("BasicRule Example")
   grammar.add_rule(rule)
   grammar.load()


The above :class:`BasicRule` example can be defined without sub-classing::

   rule = BasicRule(
       element=Repetition(
           Alternative((
               Literal("test one", value=Text("1")),
               Literal("test two", value=Text("2")),
               Literal("test three", value=Text("3")),
           )),
           1, 5)
   )


Class reference
............................................................................

"""

from six import string_types

from .rule_base  import Rule
from .elements   import ElementBase
from ..actions.action_base import ActionBase


class BasicRule(Rule):
    """
        Rule class for implementing complete or partial voice-commands
        defined using an element.

        Constructor arguments:
         - *name* (*str*) -- the rule's name
         - *element* (*Element*) -- root element for this rule
         - *exported* (boolean) -- whether the rule is exported
         - *context* (*Context*) -- context in which the rule will be active

    """

    context = None
    _default_exported = True

    def __init__(self, name=None, element=None, exported=None,
                 context=None):
        if name    is None: name    = self.__class__.__name__
        if context is None: context = self.context

        # Complex handling of exported, because of clashing use of the
        #  exported name at the class level: property & class-value.
        if exported is not None:
            pass
        elif (hasattr(self.__class__, "exported")
              and not isinstance(self.__class__.exported, property)):
            exported = self.__class__.exported
        else:
            exported = self._default_exported

        # Similar complex handling of element.
        if element is not None:
            pass
        elif (hasattr(self.__class__, "element")
              and not isinstance(self.__class__.element, property)):
            element = self.__class__.element

        # Type checking of initialization values.
        assert isinstance(name, string_types)
        assert isinstance(element, ElementBase)

        self._name     = name

        Rule.__init__(self, self._name, element, exported=exported,
                      context=context)

    def value(self, node):
        node = node.children[0]
        value = node.value()

        if hasattr(value, "copy_bind"):
            # Prepare *extras* dict for passing to _copy_bind().
            extras = {
                "_grammar":            self.grammar,
                "_rule":               self,
                "_node":               node,
            }
            element = self._element
            name = element.name
            extra_node = node.get_child_by_name(name, shallow=True)
            if extra_node:
                extras[name] = extra_node.value()
            elif element.has_default():
                extras[name] = element.default

            value = value.copy_bind(extras)

        return value

    def process_recognition(self, node):
        """
            Process a recognition of this rule.

            This method is called by the containing Grammar when this
            rule is recognized.  This method collects information about
            the recognition and then calls *self._process_recognition*.

            - *node* -- The root node of the recognition parse tree.
        """
        # Prepare *extras* dict for passing to _process_recognition().
        extras = {
            "_grammar":  self.grammar,
            "_rule":     self,
            "_node":     node,
        }
        element = self._element
        name = element.name
        extra_node = node.get_child_by_name(name, shallow=True)
        if extra_node:
            extras[name] = extra_node.value()
        elif element.has_default():
            extras[name] = element.default

        # Call the method to do the actual processing.
        self._process_recognition(node, extras)

    def _process_recognition(self, node, extras):
        """
            Default recognition processing.

            This is the method which should be overridden in most cases
            to provide derived classes with custom recognition
            processing functionality.

            This default processing method executes actions if the node's
            value is an action or a list of actions.

            - *node* -- The root node of the recognition parse tree.
            - *extras* -- A dictionary of elements from the
              extras list contained within this recognition.
              Maps element name -> element value.

        """

        value = node.value()
        if isinstance(value, (list, tuple)):
            for item in node.value():
                if isinstance(item, ActionBase):
                    item.execute(extras)
        elif isinstance(value, ActionBase):
            value.execute(extras)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

# pylint: disable=W0603
# Suppress warnings about global statements used for configuration.

# pylint: disable=E1111
# Implementations of BaseKeyboardAction return different types for events.

import io
import os
from os.path import basename

from .action_base  import DynStrActionBase
from .keyboard     import Keyboard
from .typeables    import typeables

_CONFIG_LOADED = False
UNICODE_KEYBOARD = False
HARDWARE_APPS = [
    "tvnviewer.exe", "vncviewer.exe", "mstsc.exe", "virtualbox.exe"
]
PAUSE_DEFAULT = 0.005


def load_configuration():
    """Locate and load configuration."""
    try:
        import configparser
    except ImportError:
        import ConfigParser as configparser

    global UNICODE_KEYBOARD
    global HARDWARE_APPS
    global PAUSE_DEFAULT
    global _CONFIG_LOADED

    home = os.path.expanduser("~")
    config_folder = os.path.join(home, ".dragonfly2-speech")
    config_file = "settings.cfg"
    config_path = os.path.join(config_folder, config_file)

    if not os.path.exists(config_folder):
        os.mkdir(config_folder)
    if not os.path.exists(config_path):
        with io.open(config_path, "w") as f:
            # Write the default values to the config file.
            f.write(u'[Text]\n')
            f.write(u'hardware_apps = %s\n' % "|".join(HARDWARE_APPS))
            f.write(u'unicode_keyboard = %s\n' % UNICODE_KEYBOARD)
            f.write(u'pause_default = %f\n' % PAUSE_DEFAULT)

    parser = configparser.ConfigParser()
    parser.read(config_path)
    if parser.has_option("Text", "hardware_apps"):
        HARDWARE_APPS = (parser.get("Text", "hardware_apps")
                         .lower().split("|"))
    if parser.has_option("Text", "unicode_keyboard"):
        UNICODE_KEYBOARD = parser.getboolean("Text", "unicode_keyboard")
    if parser.has_option("Text", "pause_default"):
        PAUSE_DEFAULT = parser.getfloat("Text", "pause_default")
        BaseKeyboardAction._pause_default = PAUSE_DEFAULT

    _CONFIG_LOADED = True


class BaseKeyboardAction(DynStrActionBase):
    """
        Base keystroke emulation action.

    """

    _keyboard = Keyboard()
    _pause_default = PAUSE_DEFAULT

    def __init__(self, spec=None, static=False, use_hardware=False):
        self._use_hardware = use_hardware

        # Load the Windows-only config file if necessary.
        if not _CONFIG_LOADED and os.name == "nt":
            load_configuration()

        super(BaseKeyboardAction, self).__init__(spec, static)

    def require_hardware_events(self):
        """
        Return `True` if the current context requires hardware emulation.
        """
        if self._use_hardware:
            return True

        # Load the keyboard configuration, if necessary.
        global _CONFIG_LOADED
        if not _CONFIG_LOADED:
            load_configuration()

        # Otherwise check if hardware events should be used with the current
        # foreground window.
        from dragonfly.windows import Window
        foreground_executable = basename(Window.get_foreground()
                                         .executable.lower())
        return ((not UNICODE_KEYBOARD) or
                (foreground_executable in HARDWARE_APPS))

    def _get_typeable(self, key_symbol, use_hardware):
        # Use the Typeable object for the symbol, if it exists.
        typeable = typeables.get(key_symbol)
        if typeable:
            # Update the object and return it.
            typeable.update(use_hardware)
            return typeable

        # Otherwise, get a new Typeable for the symbol, if possible.
        is_text = not use_hardware
        try:
            typeable = self._keyboard.get_typeable(key_symbol,
                                                   is_text=is_text)
        except ValueError:
            pass

        # If getting a Typeable failed, then, if it is allowed, try
        #  again with is_text=True.  On Windows, this will use Unicode
        #  events instead.
        if not (typeable or use_hardware):
            try:
                typeable = self._keyboard.get_typeable(key_symbol,
                                                       is_text=True)
            except ValueError:
                pass

        return typeable



#
# This file is part of Dragonfly.
# (c) Copyright 2018 by Dane Finlay
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
SR back-end package for the text input engine
============================================================================

"""

import logging
_log = logging.getLogger("engine.text")


# Module level singleton instance of this engine implementation.
_engine = None


def is_engine_available():
    """ Check whether the engine is available. """
    return True


def get_engine():
    """ Retrieve the back-end engine object. """
    global _engine
    if not _engine:
        from .engine import TextInputEngine
        _engine = TextInputEngine()
    return _engine



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Dictation container base class
============================================================================

This class is used to store the recognized results of dictation elements
within voice-commands.  It offers access to both the raw spoken-form words
and be formatted written-form text.

The object can be expected to behave like a string,
responding as you would expect to string methods like :meth:`replace`.
The formatted text can be retrieved using
:meth:`~DictationContainerBase.format` or simply by  calling ``str(...)``
on a dictation container object.
By default, formatting returns the words joined with
spaces, but custom formatting can be applied by calling
string methods on the :class:`Dictation` object.
A tuple of the raw  spoken words can be
retrieved using :attr:`~DictationContainerBase.words`.


String Formatting Examples
----------------------------------------------------------------------------

The following examples demonstrate dictation input can be formatted by
calling string methods on :class:`Dictation` elements.

Python example:

..  code:: python

    mapping = {
        # Define commands for writing Python methods, functions and classes.
        "method [<under>] <snaketext>":
            Text("def %(under)s%(snaketext)s(self):") + Key("left:2"),
        "function <snaketext>":
            Text("def %(snaketext)s():") + Key("left:2"),
        "classy [<classtext>]":
            Text("class %(classtext)s:") + Key("left"),

        # Define a command for accessing object members.
        "selfie [<under>] [<snaketext>]":
            Text("self.%(under)s%(snaketext)s"),
    }

    extras = [
        # Define a Dictation element that produces snake case text,
        # e.g. hello_world.
        Dictation("snaketext", default="").lower().replace(" ", "_"),

        # Define a Dictation element that produces text matching Python's
        # class casing, e.g. DictationContainer.
        Dictation("classtext", default="").title().replace(" ", ""),

        # Allow adding underscores before cased text.
        Choice("under", {"under": "_"}, default=""),
    ]

    rule = MappingRule(name="PythonExample", mapping=mapping, extras=extras)


Markdown example:

..  code:: python

    mapping = {
        # Define a command for typing Markdown headings 1 to 7 with optional
        # capitalized text.
        "heading [<num>] [<capitalised_text>]":
            Text("#")*Repeat("num") + Text(" %(capitalised_text)s"),
    }

    extras = [
        Dictation("capitalised_text", default="").capitalize(),
        IntegerRef("num", 1, 7, 1),
    ]

    rule = MappingRule(name="MdExample", mapping=mapping, extras=extras)


Camel-case example using the :meth:`Dictation.camel` method:

..  code:: python

    mapping = {
        # Define a command for typing camel-case text, e.g. helloWorld.
        "camel <camel_text>": Text(" %(camel_text)s"),
    }

    extras = [
        Dictation("camel_text", default="").camel(),
    ]

    rule = MappingRule(name="CamelExample", mapping=mapping, extras=extras)


Example using the :meth:`Dictation.apply` method for random casing:

..  code:: python

    from random import random

    def random_text(text):
        # Randomize the case of each character.
        result = ""
        for c in text:
            r = random()
            if r < 0.5:
                result += c.lower()
            else:
                result += c.upper()
        return result

    mapping = {
        "random <random_text>": Text("%(random_text)s"),
    }

    extras = [
        Dictation("random_text", default="").apply(random_text),
    ]

    rule = MappingRule(name="RandomExample", mapping=mapping, extras=extras)


Class reference
----------------------------------------------------------------------------

"""


#---------------------------------------------------------------------------
# Dictation base class -- base class for SR engine-specific containers
#  of dictated words.
import locale

from six import PY2


class DictationContainerBase(object):
    """
        Container class for dictated words as recognized by the
        :class:`Dictation` element.

        This base class implements the general functionality of dictation
        container classes.  Each supported engine should have a derived
        dictation container class which performs the actual engine-
        specific formatting of dictated text.

    """

    def __init__(self, words, methods=None):
        """
            A dictation container is created by passing it a sequence
            of words as recognized by the backend SR engine.
            Each word must be a Unicode string.

            :param words: A sequence of Unicode strings.
            :type words: sequence-of-unicode

            :param methods: Tuples describing string methods to call on the output.
            :type methods: list-of-triples

        """
        self._words = tuple(words)
        self._formatted = None
        self._methods = methods

    def __str__(self):
        if PY2:
            return self.__unicode__().encode(locale.getpreferredencoding())
        else:
            return self.__unicode__()

    def __unicode__(self):
        if self._formatted is None:
            self._formatted = self.format()
        return self._formatted

    def __repr__(self):
        message = u"%s(%s)" % (self.__class__.__name__,
                               u", ".join(self._words))
        if PY2:
            return message.encode(locale.getpreferredencoding())
        else:
            return message

    def __getattr__(self, name):
        return getattr(self.__str__(), name)

    def __add__(self, other):
        return self.__str__() + other

    def __radd__(self, other):
        return other + self.__str__()

    def __mul__(self, other):
        return self.__str__() * other

    def __rmul__(self, other):
        return self.__str__() * other

    def __getitem__(self, key):
        return self.__str__()[key]

    def __bool__(self):
        return bool(self.__str__())
    __nonzero__ = __bool__  # PY2 compatibility

    def __len__(self):
        return len(self.__str__())

    def __contains__(self, item):
        return item in self.__str__()

    @property
    def words(self):
        """ Sequence of the words forming this dictation. """
        return self._words

    def format(self):
        """ Format and return this dictation as a Unicode object. """
        return self.apply_methods(u" ".join(self._words))

    def apply_methods(self, joined_words):
        """
        Apply any string methods called on the :class:`Dictation` object to
        a given string.

        Called during :meth:`format`.
        """
        result = joined_words
        if result: # Do nothing for empty string
            for method in self._methods:
                if hasattr(result, method[0]):
                    function = getattr(result, method[0])
                    result = function(*method[1], **method[2])
                elif hasattr(self, method[0]):
                    function = getattr(self, method[0])
                    result = function(result, *method[1], **method[2])
                else:
                    raise AttributeError("'%s' is not a valid dictation or "
                                         "string method" % method[0])
        return result

    def apply(self, str_input, format_func):
        if callable(format_func):
            return format_func(str_input)
        else:
            raise TypeError("Argument passed to 'Dictation.apply' method "
                            "must be callable, taking and returning a "
                            "string.")

    def camel(self, str_input):
        def f(s):
            return s[0] + (s.title().replace(" ", "")[1:]
                    if len(s)>1 else "")
        return self.apply(str_input, f)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Recognition observer class for the Sphinx engine
============================================================================

"""

from ..base import RecObsManagerBase


class SphinxRecObsManager(RecObsManagerBase):
    """
    This class's methods are called by the engine directly, rather than
    through a grammar.
    """
    def __init__(self, engine):
        RecObsManagerBase.__init__(self, engine)

    # The following methods must be implemented by RecObsManagers.
    def _activate(self):
        pass

    def _deactivate(self):
        pass

    def notify_next_rule_part(self, words):
        """
        Notify observers that the next part of a rule has been spoken.

        This is for rules involving Dictation elements that must be spoken
        in sequence.
        """
        for observer in self._observers:
            try:
                if hasattr(observer, "on_next_rule_part"):
                    observer.on_next_rule_part(words)
            except Exception as e:
                self._log.exception("Exception during on_next_rule_part()"
                                    " method of recognition observer %s: %s"
                                    % (observer, e))



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
This file implements an interface to the Windows system clipboard.
"""

# pylint: disable=E0401
# This file imports Win32-only modules.

# pylint: disable=W0622
# Suppress warnings about redefining the built-in 'format' function.

import contextlib
import logging
import sys
import time
import threading

from six import integer_types, reraise

import pywintypes
import win32clipboard
import win32con

from .base_clipboard import BaseClipboard

#===========================================================================

class win32_clipboard_ctx(object):
    """
    Python context manager for safely opening the Windows clipboard by
    polling for access, timing out after the specified number of seconds.

    Arguments:
     - *timeout* (float, default: 0.5) -- timeout in seconds.
     - *step* (float, default: 0.001) -- number of seconds between each
       attempt to open the clipboard.

    Notes:
     - The polling is necessary because the clipboard is a shared resource
       which may be in use by another process.
     - Nested usage will not close the clipboard early.

    Use with a Python 'with' block::

       with win32_clipboard_ctx():
           # Do clipboard operation(s).
           win32clipboard.EmptyClipboard()

    """

    _ctx_data = threading.local()
    _log = logging.getLogger("clipboard")

    def __init__(self, timeout=0.5, step=0.001):
        self._timeout = float(timeout)
        self._step = float(step)

    def __enter__(self):
        timeout = time.time() + self._timeout
        success = False
        while time.time() < timeout:
            # Attempt to open the clipboard, catching Windows errors.
            try:
                win32clipboard.OpenClipboard()
                success = True
                break
            except pywintypes.error:
                # Failure. Try again after *step* seconds.
                time.sleep(self._step)

        # Try opening the clipboard one more time if it still isn't open.
        #  If this fails, then an error will be raised this time.
        if not success:
            win32clipboard.OpenClipboard()

        # The clipboard is open now.  Increment our thread-local counter to
        #  keep track of nested usage.
        ctx_data = self._ctx_data
        ctx_data.count = getattr(ctx_data, "count", 0) + 1

    def __exit__(self, exc_type, exc_val, exc_tb):
        # The clipboard is (presumably) still open.  Decrement our thread-
        #  local counter and close the clipboard if it is at 0 again.
        ctx_data = self._ctx_data
        ctx_data.count -= 1
        if ctx_data.count == 0:
            win32clipboard.CloseClipboard()


class Win32Clipboard(BaseClipboard):
    """
    Class for interacting with the Windows system clipboard.

    This is Dragonfly's default clipboard class on Windows.

    """

    #-----------------------------------------------------------------------

    format_text      = win32con.CF_TEXT
    format_oemtext   = win32con.CF_OEMTEXT
    format_unicode   = win32con.CF_UNICODETEXT
    format_locale    = win32con.CF_LOCALE
    format_hdrop     = win32con.CF_HDROP
    format_names = {
        format_text:     "text",
        format_oemtext:  "oemtext",
        format_unicode:  "unicode",
        format_locale:   "locale",
        format_hdrop:    "hdrop",
    }

    @classmethod
    def get_system_text(cls):
        with win32_clipboard_ctx():
            try:
                content = win32clipboard.GetClipboardData(
                    cls.format_unicode)
                if not content:
                    content = win32clipboard.GetClipboardData(
                        cls.format_text)
            except (TypeError, pywintypes.error):
                content = u""
        return content

    @classmethod
    def set_system_text(cls, content):
        # If *content* is None, clear the clipboard and return early.
        if content is None:
            cls.clear_clipboard()
            return

        # Otherwise, convert *content* and set the system clipboard data.
        content = cls.convert_format_content(cls.format_unicode, content)
        with win32_clipboard_ctx():
            win32clipboard.EmptyClipboard()
            win32clipboard.SetClipboardData(cls.format_unicode, content)

    @classmethod
    def clear_clipboard(cls):
        with win32_clipboard_ctx():
            win32clipboard.EmptyClipboard()

    @classmethod
    def _wait_for_change(cls, timeout, step, formats, initial_clipboard,
                         seq_no):
        # This method determines if the system clipboard has changed by
        #  repeatedly checking the current sequence number.  Contents are
        #  compared if specific clipboard formats are given.
        timeout = time.time() + float(timeout)
        step = float(step)
        if isinstance(formats, integer_types):
            formats = (formats,)
        elif formats:
            for format in formats:
                if not isinstance(format, integer_types):
                    raise TypeError("Invalid clipboard format: %r"
                                    % format)
        clipboard2 = cls() if formats or initial_clipboard else None
        result = False
        while time.time() < timeout:
            seq_no_change = (win32clipboard.GetClipboardSequenceNumber()
                             != seq_no)
            if seq_no_change and initial_clipboard:
                # Check if the content of any relevant format has changed.
                clipboard2.copy_from_system()
                formats_to_compare = formats if formats else "all"
                result = cls._clipboard_formats_changed(formats_to_compare,
                                                        initial_clipboard,
                                                        clipboard2)

                # Reset the sequence number if the clipboard change is not
                #  related.
                if not result:
                    seq_no = win32clipboard.GetClipboardSequenceNumber()
            elif seq_no_change:
                result = True

            if result:
                break

            # Failure. Try again after *step* seconds.
            time.sleep(step)
        return result

    @classmethod
    def wait_for_change(cls, timeout, step=0.001, formats=None,
                        initial_clipboard=None):
        # Save the current clipboard sequence number and clipboard contents,
        #  as necessary. The latter is not required unless formats are
        #  specified.
        seq_no = win32clipboard.GetClipboardSequenceNumber()
        if formats and not initial_clipboard:
            initial_clipboard = cls(from_system=True)
        return cls._wait_for_change(timeout, step, formats,
                                    initial_clipboard, seq_no)

    @classmethod
    @contextlib.contextmanager
    def synchronized_changes(cls, timeout, step=0.001, formats=None,
                             initial_clipboard=None):
        seq_no = win32clipboard.GetClipboardSequenceNumber()
        if formats and not initial_clipboard:
            initial_clipboard = cls(from_system=True)
        try:
            # Yield for clipboard operations.
            yield
        finally:
            # Wait for the system clipboard to change.
            cls._wait_for_change(timeout, step, formats, initial_clipboard,
                                 seq_no)

    #-----------------------------------------------------------------------

    def _update_format_names(self, formats):
        # Retrieve and store the names of any registered clipboard formats
        #  in format_names, if necessary.
        for format in formats:
            if format in self.format_names:
                continue

            # Registered clipboard formats are identified by values in
            #  the range 0xC000 through 0xFFFF.
            if 0xC000 <= format <= 0xFFFF:
                format_name = win32clipboard.GetClipboardFormatName(format)
                self.format_names[format] = format_name

    def copy_from_system(self, formats=None, clear=False):
        with win32_clipboard_ctx():
            # Determine which formats to retrieve.
            caller_specified_formats = bool(formats)
            if not formats:
                format = 0
                formats = []
                while 1:
                    format = win32clipboard.EnumClipboardFormats(format)
                    if not format:
                        break
                    formats.append(format)
            elif isinstance(formats, integer_types):
                formats = (formats,)

            # Verify that the given formats are valid.
            for format in formats:
                if not isinstance(format, integer_types):
                    raise TypeError("Invalid clipboard format: %r"
                                    % format)

            # Update clipboard format names.
            self._update_format_names(formats)

            # Retrieve Windows system clipboard content.
            contents = {}
            for format in formats:
                # Allow GetClipboardData() to raise Windows API errors if
                #  the caller specified one or more formats.  If the
                #  function raises an error at this point, it is likely the
                #  format is not retrievable and the caller will want to
                #  know, since they explicitly asked for this format.
                if caller_specified_formats:
                    content = win32clipboard.GetClipboardData(format)
                    contents[format] = content
                    continue

                # Otherwise, catch and log Windows API errors.  We assume
                #  here that the caller is not much interested in
                #  irretrievable formats.
                try:
                    content = win32clipboard.GetClipboardData(format)
                    contents[format] = content
                except pywintypes.error as err:
                    format_repr = self.format_names.get(format, format)
                    self._log.warning("Could not retrieve clipboard data "
                                      "for format %r: %s", format_repr, err)

            self._contents = contents

            # Clear the system clipboard, if requested, and close it.
            if clear:
                win32clipboard.EmptyClipboard()

    def copy_to_system(self, clear=True):
        with win32_clipboard_ctx():
            # Clear the system clipboard, if requested.
            if clear:
                win32clipboard.EmptyClipboard()

            # Transfer content to Windows system clipboard.
            for format, content in self._contents.items():
                win32clipboard.SetClipboardData(format, content)
